from typing import Type

from stilus.functions.rgba import rgba
from stilus.nodes.color import Color, HSLA
from stilus.nodes.unit import Unit


def alpha(color: Type[Color], value=None, evaluator=None):
    if isinstance(color, HSLA):
        color = color.rgba()
    if value:
        return rgba(
            Unit(color.r), Unit(color.g), Unit(color.b), value, evaluator
        )
    return Unit(color.a, "")



import json

from .atrule import Atrule


class Supports(Atrule):
    def __init__(self, condition, lineno=1, column=1):
        super().__init__("supports", lineno=lineno, column=column)
        self.condition = condition
        self.block = None

    def __str__(self):
        return f"@supports {self.condition}"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.node_name, self.condition

    def __eq__(self, other):
        if isinstance(other, Supports):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.__key())

    def clone(self, parent=None, node=None):
        clone = Supports(None, lineno=self.lineno, column=self.column)
        clone.condition = self.condition.clone(parent, clone)
        clone.block = self.block.clone(parent, clone)
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Supports",
                "condition": self.condition,
                "block": self.block,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



import json

from stilus import utils
from .boolean import Boolean
from .node import Node


class Expression(Node):
    def __init__(self, is_list=False, preserve=False, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.nodes = []
        self.is_list = is_list
        self.preserve = preserve

    def __str__(self):
        """
        :return: "(<a> <b> <c>)" or "(<a>, <b>, <c>)" if the expression
        represents a list.
        """
        separator = " "
        if self.is_list:
            separator = ", "
        return "(" + separator.join(map(str, self.nodes)) + ")"

    def __repr__(self):
        return [n.__str__() for n in self.nodes]

    def __key(self):
        return self.nodes

    def __eq__(self, other):
        if isinstance(other, Expression):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return id(self)

    def hash(self):
        hashes = [str(node.hash()) for node in self.nodes]
        return "::".join(hashes)

    def __len__(self):
        return len(self.nodes)

    def __iter__(self):
        return iter(self.nodes)

    def is_empty(self):
        return len(self.nodes) == 0

    def first(self):
        if not self.is_empty():
            return self.nodes[0].first()
        else:
            from .null import null

            return null

    def clone(self, parent=None, node=None):
        clone = Expression(
            self.is_list, lineno=self.lineno, column=self.column
        )
        clone.preserve = self.preserve
        clone.filename = self.filename
        for something in self.nodes:
            clone.append(something.clone(parent, clone))
        return clone

    def to_boolean(self):
        if len(self.nodes) > 1:
            return Boolean(True)
        return self.first().to_boolean()

    def to_json(self):
        return json.dumps(
            {
                "__type": "Expression",
                "isList": self.is_list,
                "preserve": self.preserve,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
                "nodes": self.nodes,
            }
        )

    def operate(self, op, right, value=None):  # noqa: C901
        if op == "[]=":
            nodes = utils.unwrap(right).nodes
            value = utils.unwrap(value)
            for node in nodes:
                if node.node_name == "unit":
                    index = utils.get_value(node)
                    last = len(self.nodes) - 1
                    for i in range(last, index):
                        from stilus.nodes.null import null

                        self.nodes.append(null)
                    self.nodes[index] = value
                elif node.string:
                    if (
                        not self.is_empty()
                        and self.nodes[0].node_name == "objectnode"
                    ):
                        self.nodes[0].set(node.string, value.clone())
            return value
        elif op == "[]":
            expression = Expression()
            values = utils.unwrap(self).nodes
            nodes = utils.unwrap(right).nodes
            for node in nodes:
                n = None
                if node.node_name == "unit":
                    index = utils.get_value(node)
                    try:
                        n = values[index]
                    except IndexError:
                        pass
                elif len(values) > 0 and values[0].node_name == "objectnode":
                    n = values[0].get(node.string)
                if n:
                    expression.append(n)
            from .null import null

            return null if expression.is_empty() else utils.unwrap(expression)
        elif op == "||":
            return self if self.to_boolean().value is True else right
        elif op == "in":
            return super().operate(op, right, self.value)
        elif op == "!=":
            return self.operate("==", right, value).negate()
        elif op == "==":
            right = right.to_expression()
            if len(self.nodes) != len(right.nodes):
                return Boolean(False)
            for i in range(len(self.nodes)):
                a = self.nodes[i]
                b = right.nodes[i]
                if a.operate(op, b).is_true():
                    continue
                return Boolean(False)
            return Boolean(True)
        else:
            return self.first().operate(op, right, value)

    def append(self, node):
        self.nodes.append(node)



import inspect
import logging
from pathlib import Path
from urllib.parse import urlparse

from stilus import utils
from stilus.nodes.literal import Literal
from stilus.visitor.compiler import Compiler

log = logging.getLogger(__name__)


# todo: check evaluator; could be stilus?
# fixme: rewrite this
def resolver(url, options=None, evaluator=None):
    frame = inspect.stack()[1]
    module = inspect.getmodule(frame[0])
    log.debug(
        f"Resolver called by {module.__file__} " f"({frame[0].f_lineno})."
    )
    log.debug(f'Functions: {evaluator.options["functions"]}.')
    if not options:
        options = {}

    original = url  # super ugly

    # compile the urls nodes and create a url from the result
    compiler = Compiler(url, options)
    filename = url.filename
    compiler.is_url = True
    url = urlparse(compiler.visit(url))

    # fixme: dirty hack
    if url.geturl() == "" and f"{original}" in ["'#'", "('#')"]:
        literal = Literal('url("#")')
    else:
        # regular call
        literal = Literal(f'url("{url.geturl()}")')
    path = url.path
    dest = options.get("dest", "")
    tail = ""

    # absolute or hash
    if url.scheme or not path or "/" == path[0]:
        return literal

    # check that a file exists
    if options.get("nocheck", True):
        _paths = options.get("paths", [])
        _paths.extend(evaluator.paths)
        path = utils.lookup(path, _paths)
        if not path:
            return literal
        else:
            path = Path(path)

    if evaluator.include_css and path.suffix == ".css":
        return Literal(url.geturl())

    if url.query or url.fragment:
        # fixme: extend url with a sep!
        if "#" in f"{original}":
            tail += "#"
        else:
            tail += "?"
    if url.query:
        tail += url.query
    if url.fragment:
        tail += url.fragment

    if dest and dest.suffix == ".css":
        dest = dest.parent

    if dest:
        first = dest.parents[1]
    else:
        first = Path(evaluator.filename).parent
    if options.get("nocheck", False):
        other = Path(filename).parent
    else:
        other = path

    res = other.relative_to(first.resolve())

    # use the first path of the options['paths'] list as cwd
    cwd = Path(evaluator.options.get("paths", ["."])[0])
    try:
        res = f"{res.resolve().relative_to(cwd)}{tail}"
    except ValueError:
        res = f"{res}{tail}"

    # todo: handle windows separators?

    return Literal(f'url("{res}")')


def get_resolver():
    return resolver



from stilus.exceptions import StilusError
from stilus.utils import assert_color, assert_string, assert_type


def adjust(color, prop, amount, evaluator=None):
    assert_color(color, "color")
    assert_string(prop, "prop")
    assert_type(amount, "unit", "amount")
    hsl = color.hsla().clone()
    if not hasattr(hsl, prop.string):
        raise StilusError("Invalid adjustment property.")
    value = amount.value
    if amount.type == "%":
        if prop.string == "lightness" and value > 0:
            value = (100 - hsl.lightness) * value / 100
        else:
            value = getattr(hsl, prop.string) * (value / 100)
    setattr(hsl, prop.string, getattr(hsl, prop.string) + value)
    return hsl.rgba()



from stilus.nodes.string import String
from stilus.nodes.ident import Ident
from stilus.utils import assert_string


def replace(pattern, replacement, value, evaluator=None):
    assert_string(pattern, "pattern")
    assert_string(replacement, "replacement")
    assert_string(value, "value")
    result = value.string.replace(pattern.string, replacement.string)
    if isinstance(value, Ident):
        return Ident(result)
    else:
        return String(result)



import json

from .node import Node


class Feature(Node):
    def __init__(self, segments=None, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.segments = segments
        self.expr = None
        self.name = None

    def __str__(self):
        if self.expr:
            s = ""
            for segment in self.segments:
                s += str(segment)
            return f"({s}: {self.expr})"
        else:
            return "".join(self.segments)

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return self.segments, self.expr

    def __eq__(self, other):
        if isinstance(other, Feature):
            return self.__key() == other.__key()
        return False

    def clone(self, parent=None, node=None):
        clone = Feature(lineno=self.lineno, column=self.column)
        clone.segments = [node.clone(parent, clone) for node in self.segments]
        if self.expr:
            clone.expr = self.expr.clone(parent, clone)
        clone.name = self.name
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Feature",
                "segment": self.segments,
                "expr": self.expr,
                "name": self.name,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



from stilus.nodes.node import Node

import logging

log = logging.getLogger(__name__)


class Visitor:
    def __init__(self, root):
        self.root = root

    def visit(self, node: Node):
        """Visit a method based on the node's type.  For example when the
        bode has a node_name == 'block', the method visit_block(node)
        will be called.
        :param node: The node who's type will be used to visit a method.
        :return: the node if its type does not have a method, or the
        return value of the visited method.
        """

        method = None
        if hasattr(node, "node_name"):
            method = f"visit_{node.node_name}"
            # log.info(f'Visiting {method}({node.node_name}).')
            if self.callable(method):
                log.debug(f"{method} is callable. [{node.node_name}]")
                return getattr(self, method)(node)

            # debug else
            if log.isEnabledFor(logging.DEBUG) and not self.callable(method):
                log.debug(f"{method} is NOT a callable! [{node.node_name}]")

        # debug else
        if log.isEnabledFor(logging.DEBUG) and not hasattr(node, "node_name"):
            log.debug(f"{type(node)} has no node_name attribute!")

        if method:
            log.info(
                f"Returning {type(node)} from " f"{method}({node.node_name})."
            )
        else:
            log.info(f"Skipping, since no visiting method found for {node}.")
        return node

    def callable(self, method: str):
        methods = [f for f in dir(self) if callable(getattr(self, f))]
        if method in methods:
            return True
        return False



from stilus.nodes.color import Color
from stilus.nodes.unit import Unit
from stilus.utils import assert_color


def luminosity(color: Color, evaluator=None):
    assert_color(color)
    color = color.rgba()

    def process_channel(channel):
        channel = channel / 255
        if 0.03928 > channel:
            return channel / 12.92
        else:
            return pow((channel + 0.055) / 1.055, 2.4)

    return Unit(
        0.2126 * process_channel(color.r)
        + 0.7152 * process_channel(color.g)
        + 0.0722 * process_channel(color.b)
    )



from stilus.nodes import null
from stilus.utils import assert_type

import logging

log = logging.getLogger(__name__)


def warn(msg, evaluator=None):
    assert_type(msg, "string", "msg")
    print(f"Warning: {msg.value}")
    log.warning(f"Warning: {msg.value}")
    return null



import json

from .block import Block
from .node import Node
from .selector import Selector


class Group(Node):
    def __init__(self, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.nodes = []
        self.extends = []

    def __str__(self):
        return f"{self.nodes}:{self.extends}"

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        return id(self) == id(other)

    def append(self, selector: Selector):
        self.nodes.append(selector)

    @property
    def block(self):
        if self.nodes[0]:
            return self.nodes[0].block
        else:
            return None

    @block.setter
    def block(self, block: Block):
        for node in self.nodes:
            node.block = block

    def has_only_placeholders(self):
        return all(s.is_placeholder() for s in self.nodes)

    def clone(self, parent=None, noe=None):
        clone = Group(lineno=self.lineno, column=self.column)
        clone.filename = self.filename
        clone.nodes = [node.clone(parent, clone) for node in self.nodes]
        clone.block = self.block.clone(parent, clone)
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Group",
                "nodes": self.nodes,
                "block": self.block,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



from stilus.functions.rgba import rgba
from stilus.nodes.color import RGBA
from stilus.nodes.unit import Unit
from stilus.utils import assert_color


def rgb(red, green, blue, evaluator=None):
    if red and (green is blue is None):
        assert_color(red)
        color = red.rgba()
        return RGBA(color.r, color.g, color.b, 1)
    return rgba(red, green, blue, Unit(1), evaluator)



from stilus.exceptions import StilusError
from stilus.utils import assert_type


def error(msg, evaluator=None):
    assert_type(msg, "string", "msg")
    err = StilusError(msg.value)
    err.from_stilus = True
    raise err



from stilus.nodes.string import String


def current_media(*args, evaluator=None):
    def look_for_media(node):
        if node.node_name == "media":
            node.value = evaluator.visit(node.value)
            return str(node)
        else:
            block = node.block
            if block and block.parent and hasattr(block.parent, "node"):
                return look_for_media(block.parent.node)
        return None

    media = look_for_media(evaluator.closest_block().node)
    if media:
        return String(media)
    else:
        return String("")



import json

from .arguments import Arguments
from .node import Node


class Call(Node):
    def __init__(
        self, function_name, args: Arguments = None, lineno=1, column=1
    ):
        super().__init__(lineno=lineno, column=column)
        self.function_name = function_name
        self.block = None
        self.args = args

    def __str__(self):
        arguments = [str(arg)[1:-1] for arg in self.args]
        return f'{self.function_name}({", ".join(arguments)})'

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.node_name, self.function_name, self.args

    def __eq__(self, other):
        if isinstance(other, Call):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.value)

    def clone(self, parent=None, clone=None):
        clone = Call(self.function_name)
        clone.args = self.args.clone(parent, clone)
        if self.block:
            clone.block = self.block.clone(parent, clone)
        clone.lineno = self.lineno
        clone.column = self.column
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Call",
                "node_name": self.node_name,
                "args": self.args,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



import json

from .node import Node


class QueryList(Node):
    def __init__(self, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.nodes = []

    def __str__(self):
        lst = []
        for node in self.nodes:
            lst.append(str(node))
        return f'({", ".join(lst)})'

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return self.nodes

    def __eq__(self, other):
        if isinstance(other, QueryList):
            return self.__key() == other.__key()
        return False

    def merge(self, other):
        list = QueryList()
        for query in self.nodes:
            for i, other_node in enumerate(other.nodes):
                merged = query.merge(other_node)
                if merged:
                    list.append(merged)
        return list

    def append(self, node):
        self.nodes.append(node)

    def clone(self, parent=None, node=None):
        clone = QueryList(lineno=self.lineno, column=self.column)
        clone.filename = self.filename
        clone.nodes = [node.clone(parent, clone) for node in self.nodes]
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "QueryList",
                "nodes": self.nodes,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



import json

from .node import Node


class Null(Node):
    def __init__(self):
        super().__init__()
        self.lineno = 0
        self.column = 0

    def is_null(self):
        return True

    def __str__(self):
        return "null"

    def __repr__(self):
        return str(self)

    def __bool__(self):
        return False

    def __eq__(self, other):
        if isinstance(other, Null):
            return True
        return False

    def __hash__(self):
        return hash(None)

    def hash(self):
        return None

    def clone(self, parent=None, node=None):
        return self

    def to_expression(self):
        return super().to_expression()

    def to_boolean(self):
        from .boolean import false

        return false

    def to_json(self):
        return json.dumps(
            {
                "__type": "Null",
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )


null = Null()



from stilus.utils import assert_string, assert_type


def remove(object, key, evaluator=None):
    assert_type(object, "objectnode", "object")
    assert_string(key, "key")
    if object.has(key.string):
        del object.values[key.string]
    return object



from stilus.utils import unwrap


def unshift(expr, *args, evaluator=None):
    expr = unwrap(expr)
    for arg in args:
        a = unwrap(arg)
        expr.nodes.insert(0, a)
    return len(expr.nodes)



from stilus.nodes.null import null
from stilus.utils import unwrap


def p(*args, evaluator=None):
    for arg in args:
        expr = unwrap(arg)
        if not expr:
            return None
        e = str(expr)
        if e.startswith("("):
            e = e[1:]
        if e.endswith(")"):
            e = e[:-1]
        print(f"inspect: {e}")
    return null



from .parser import Parser
from .utils import coerce
from .visitor.compiler import Compiler
from .visitor.evaluator import Evaluator
from .visitor.normalizer import Normalizer


class Renderer:
    def __init__(self, s: str, options: dict):
        self.s = s
        self.options = options
        self.options["globals"] = options.get("globals", {})
        self.options["functions"] = options.get("functions", {})
        self.options["use"] = options.get("use", [])
        if not isinstance(self.options["use"], list):
            self.options["use"] = list(self.options["use"])
        # todo: fix me and use pathlib
        from os.path import realpath, dirname, join

        self.options["imports"] = [
            join(dirname(realpath(__file__)), "functions")
        ]
        self.options["paths"] = options.get("paths", [])
        self.options["filename"] = options.get("filename", "stilus")
        self.options["Evaluator"] = options.get("Evaluator", Evaluator)

    def render(self):

        # todo: use plugin

        self.parser = Parser(self.s, self.options)
        ast = self.parser.parse()

        self.evaluator = Evaluator(ast, self.parser, self.options)
        ast = self.evaluator.evaluate()

        self.normalizer = Normalizer(ast, self.parser, self.options)
        ast = self.normalizer.normalize()

        self.compiler = Compiler(ast, self.options)
        ast = self.compiler.compile()

        # todo: expose sourcemap

        # todo: handle end event

        return ast

    def include(self, path):
        self.options["paths"].append(path)
        return self

    def define(self, name, fn, raw=True, options=None):
        fn = coerce(fn, raw)
        if hasattr(fn, "node_name"):
            self.options["globals"][name] = fn
            return self

        self.options["functions"][name] = fn
        if raw:
            fn.raw = raw
            # todo: add it to the raw_bifs here?
        if options:
            fn.options = options
        return self

    def do_import(self, file):
        self.options["imports"].append(file)
        return self


def render(s, options) -> str:
    return Renderer(s, options).render()



from stilus.nodes.ident import Ident
from stilus.nodes.null import null
from stilus.utils import assert_type, unwrap


def define(name, expr, common=None, evaluator=None):
    assert_type(name, "string", "name")
    expr = unwrap(expr)
    scope = evaluator.get_current_scope()
    if common and common.to_boolean().is_true():
        scope = evaluator.common.scope()
    node = Ident(name.value, expr)
    scope.add(node)
    return null



import re
from collections import defaultdict

from stilus import utils
from stilus.nodes.block import Block
from stilus.nodes.group import Group
from stilus.nodes.literal import Literal
from stilus.nodes.node import Node
from stilus.nodes.null import null
from stilus.nodes.root import Root
from stilus.nodes.selector import Selector
from stilus.visitor.visitor import Visitor

import logging

log = logging.getLogger(__name__)


class Normalizer(Visitor):
    def __init__(self, root, parser, options):
        super().__init__(root)
        self.parser = parser
        self.hoist = options.get("hoist atrules", False)
        self.stack = []
        self.selector_map = defaultdict(list)
        self.imports = []
        self.charset = None
        self.root_index = 0

    def normalize(self):
        ret = self.visit(self.root)
        if self.hoist:
            # hoist @import
            if len(self.imports) > 0:
                self.imports.extend(ret.nodes)
                ret.nodes = self.imports
            # hoist @charset
            if self.charset:
                if not ret.nodes:
                    ret.nodes = []
                charsets = [self.charset]
                charsets.extend(ret.nodes)
                ret.nodes = charsets
        return ret

    def bubble(self, node: Node):
        props = []
        others = []

        def filter_props(block):
            for node in block.nodes:
                node = self.visit(node)
                if node.node_name == "property":
                    props.append(node)
                elif node.node_name == "block":
                    filter_props(node)
                else:
                    others.append(node)

        filter_props(node.block)

        if props:
            selector = Selector([Literal("&")])
            selector.lineno = node.lineno
            selector.column = node.column
            selector.filename = node.filename
            selector.value = "&"

            group = Group(lineno=self.parser.lineno, column=self.parser.column)
            group.lineno = node.lineno
            group.column = node.column
            group.filename = node.filename

            block = Block(node.block, group)
            block.lineno = node.lineno
            block.column = node.column
            block.filename = node.filename

            for prop in props:
                block.append(prop)

            group.append(selector)
            group.block = block

            node.block.nodes = []
            node.block.append(group)
            for other in others:
                node.block.append(other)

            group = self.closest_group(node.block)
            if group:
                node.group = group.clone()

            node.bubbled = True

    def closest_group(self, block: Block) -> Group:
        parent = block.parent
        while parent and hasattr(parent, "node") and parent.node:
            node = parent.node
            if node.node_name == "group":
                return node
            if not hasattr(node, "block") or not node.block:
                parent = False
            elif not node.block.parent:
                parent = False
            else:
                parent = node.block.parent

    def visit_root(self, block):
        ret = Root()
        for i, node in enumerate(block.nodes):
            if node.node_name in [
                "null",
                "expression",
                "function",
                "unit",
                "atblock",
            ]:
                continue
            else:
                self.root_index = i
                ret.append(self.visit(node))
        return ret

    def visit_property(self, property):
        self.visit(property.expr)
        return property

    def visit_expression(self, expression):
        def handle(node):
            # returns `block` literal if mixin's block
            # is used as part of a property value
            if node.node_name == "block":
                literal = Literal("block")
                literal.lineno = node.lineno
                literal.column = node.column
                return literal
            else:
                return node

        expression.nodes = [handle(node) for node in expression.nodes]
        return expression

    def visit_block(self, block):
        if block.has_properties():
            i = 0
            while i < len(block.nodes):
                node = block.nodes[i]
                if node.node_name in [
                    "null",
                    "expression",
                    "function",
                    "group",
                    "unit",
                    "atblock",
                ]:
                    i += 1
                    continue
                else:
                    block.nodes[i] = self.visit(node)
                    i += 1

        # nesting
        i = 0
        while i < len(block.nodes):
            node = block.nodes[i]
            block.nodes[i] = self.visit(node)
            i += 1

        return block

    def visit_group(self, group):
        stack = self.stack
        normalized = []
        # normalize interpolated selectors with a comma
        for selector in group.nodes:
            if selector.value:
                # do nothing
                if "," not in selector.value:
                    normalized = group.nodes
                    break
                # replace '\,' with ','
                if "\\" in selector.value:
                    selector.value = re.sub(r"\\,", ",", selector.value)
                    normalized = group.nodes
                    break
                parts = selector.value.split(",")
                root = True if selector.value[0] == "/" else False
                for i, part in enumerate(parts):
                    part = part.strip()
                    if root and "&" not in part:
                        part = "/" + part
                    s = Selector([Literal(part)])
                    s.value = part
                    s.block = group.block
                    normalized.append(s)
        group.nodes = normalized
        stack.append(normalized)

        selectors = utils.compile_selectors(stack, True)

        # add selectors for lookup
        for selector in selectors:
            self.selector_map[selector].append(group)

        # extensions
        self.extend(group, selectors)

        stack.pop()
        return group

    def visit_function(self, function):
        return null

    def visit_media(self, media):
        medias = []
        group = self.closest_group(media.block)

        def merge_queries(block):
            for i, node in enumerate(block.nodes):
                if node.node_name == "media":
                    # fixme: what does this do?
                    node.value = media.value.merge(node.value)
                    medias.append(node)
                    block.nodes[i] = null
                elif node.node_name == "block":
                    merge_queries(node)
                else:
                    if (
                        hasattr(node, "block")
                        and node.block
                        and node.block.nodes
                    ):
                        merge_queries(node.block)

        merge_queries(media.block)
        self.bubble(media)

        index = 0
        while index < len(medias):
            node = medias[index]
            if group:
                group.block.append(node)
            else:
                self.root_index += 1
                self.root.nodes.insert(self.root_index, node)
            node = self.visit(node)
            parent = node.block.parent
            if node.bubbled and (
                not group or parent.node.node_name == "group"
            ):
                node.group.block = node.block.nodes[0].block
                node.block.nodes[0] = node.group
            index += 1

        return media

    def visit_supports(self, node):
        self.bubble(node)
        return node

    def visit_atrule(self, node):
        if node.block:
            node.block = self.visit(node.block)
        return node

    def visit_keyframes(self, node):
        node.frames = len(
            [
                n
                for n in node.block.nodes
                if hasattr(n, "block") and n.block and n.block.has_properties()
            ]
        )
        return node

    def visit_import(self, node):
        self.imports.append(node)
        return null if self.hoist else node

    def visit_charset(self, node):
        self.charset = node
        return null if self.hoist else node

    def extend(self, group: Group, selectors):
        selector_map = self.selector_map
        parent = self.closest_group(group.block)

        for extend in group.extends:
            groups = selector_map[extend["selector"]]
            if not groups:
                if extend["optional"]:
                    return
                err = TypeError(f'Failed to @extend "{extend["selector"]}"')
                err.lineno = extend["lineno"]
                err.column = extend["column"]
                raise TypeError
            for selector in selectors:
                node = Selector()
                node.value = selector
                node.inherits = False
                for g in groups:
                    # prevent recursive extend
                    if not parent or parent != g:
                        self.extend(g, selectors)
                    g.append(node)

        group.block = self.visit(group.block)



import json

from .node import Node


class Charset(Node):
    def __init__(self, value, lineno=1, column=1):
        super().__init__(value, lineno=lineno, column=column)

    def __str__(self):
        return f"@charset {self.value}"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.node_name, self.value

    def __eq__(self, other):
        if isinstance(other, Charset):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.__key())

    def to_json(self):
        return json.dumps(
            {
                "__type": "Charset",
                "val": self.value,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



import json

from .node import Node


class Member(Node):
    def __init__(self, left, right, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.right = right
        self.left = left

    def __str__(self):
        return f"{self.left} = {self.right}"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.node_name, self.left, self.right

    def __eq__(self, other):
        if isinstance(other, Member):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.__key())

    def clone(self, parent=None, node=None):
        clone = Member(None, None, lineno=self.lineno, column=self.column)
        clone.left = self.left.clone(parent, clone)
        clone.right = self.right.clone(parent, clone)
        if self.value:
            clone.value = self.value.clone(parent, clone)
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Member",
                "left": self.left,
                "right": self.right,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



import json

from .node import Node


class Boolean(Node):
    def __init__(self, value: bool, lineno=1, column=1):
        from .null import null

        if value == null:
            value = False
        super().__init__(bool(value), lineno=lineno, column=column)

    def __str__(self):
        return str(self.value).lower()

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if isinstance(other, Boolean):
            return self.value == other.value
        return False

    def __hash__(self):
        # todo: fix this!
        return hash(self.value)

    def is_true(self):
        return self.value is True

    def is_false(self):
        return self.value is False

    def negate(self):
        return Boolean(not self.value)

    def to_json(self):
        return json.dumps({"__type": "Boolean", "value": self.value})

    def to_boolean(self):
        return self


true = Boolean(True)
false = Boolean(False)



import json

from .node import Node


class UnaryOp(Node):
    def __init__(self, op, expr, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.op = op
        self.expr = expr

    def __str__(self):
        return f"({self.op},{self.expr})"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.op, self.expr

    def __eq__(self, other):
        if isinstance(other, UnaryOp):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.__key())

    def clone(self, parent=None, node=None):
        clone = UnaryOp(self.op, None, lineno=self.lineno, column=self.column)
        clone.expr = self.expr.clone(parent, node)
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "UnaryOp",
                "op": self.op,
                "expr": self.expr,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



from stilus.nodes.literal import Literal
from stilus.nodes.null import null
from stilus.utils import assert_string, unwrap
from stilus.visitor.compiler import Compiler


def s(fmt, *args, evaluator=None):
    options = evaluator.options if evaluator else {}
    fmt = unwrap(fmt).nodes[0]
    assert_string(fmt, "string")
    result = fmt.string
    results = []
    for arg in args:
        from stilus.visitor.evaluator import Evaluator

        if not isinstance(arg, Evaluator):
            results.append(Compiler(arg, options).compile())
    for r in results:
        result = result.replace("%s", r, 1)
    # add nulls for missed %s elements
    c = Compiler(null, options).compile()
    result = result.replace("%s", c)
    return Literal(result)



import base64
import logging
import re
import urllib
from pathlib import Path
from urllib.parse import urlparse

from stilus import utils
from stilus.nodes.literal import Literal
from stilus.visitor.compiler import Compiler

default_mimes = {
    ".gif": "image/gif",
    ".png": "image/png",
    ".jpg": "image/jpeg",
    ".jpeg": "image/jpeg",
    ".svg": "image/svg+xml",
    ".webp": "image/webp",
    ".ttf": "application/x-font-ttf",
    ".eot": "application/vnd.ms-fontobject",
    ".woff": "application/font-woff",
    ".woff2": "application/font-woff2",
}


encoding_types = {"BASE_64": "base64", "UTF8": "UTF-8"}


# todo: rewrite this; this is terrible!
# todo: this should be linked with the resolver?
# note: this is called by the 'embedurl' builtin function.
# todo: docstring this!
def url(url, enc=None, evaluator=None):
    _paths = evaluator.options.get("paths", [])
    size_limit = evaluator.options.get("limit", 30000)
    mimes = evaluator.options.get("mimes", default_mimes)

    def fn(url, enc=None):
        compiler = Compiler(url, {})
        url = compiler.visit(url)

        # parse the url
        url = urlparse(url[1:-1])
        ext = Path(url.path).suffix
        mime = None
        if ext and ext in mimes:
            mime = mimes[ext]
        hash = ""
        if url.fragment:
            hash = f"#{url.fragment}"
        literal = Literal(f'url("{url.geturl()}")')

        # not mime or absolute
        if not mime or url.scheme:
            return literal

        # lookup
        found = utils.lookup(url.path, _paths)
        if not found:
            # todo: add event management
            logging.warning(
                f"File not found; File {literal} could not be "
                f"found, literal url retained!"
            )
            return literal

        # read the url as a binary
        buf = open(found, "rb").read()

        # too large?
        if size_limit and len(buf) > size_limit:
            return literal

        if enc and enc.first().value.lower() == "utf8":
            encoding = "charset=utf-8"
            buf = re.sub(r"\s+", " ", buf.decode("utf-8"))
            result = urllib.parse.quote(buf, safe=" ?=:/").strip()
        else:
            encoding = "base64"
            result = f'{base64.b64encode(buf).decode("utf-8")}{hash}'

        return Literal(f'url("data:{mime};{encoding},{result}")')

    return fn(url, enc)



import json

from .block import Block
from .node import Node


class Root(Node):
    def __init__(self, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.nodes = []

    def append(self, node):
        self.nodes.append(node)

    def unshift(self, node):
        self.nodes.insert(0, node)

    def __key(self):
        return self.nodes

    def __eq__(self, other):
        if isinstance(other, Root):
            return id(self) == id(other)
        return False

    def __hash__(self):
        return hash(self.__key())

    def __str__(self):
        return f"[Root]"

    def __repr__(self):
        return self.__str__()

    def clone(self, parent=None, node=None):
        clone = Root(lineno=self.lineno, column=self.column)
        clone.filename = self.filename
        clone.nodes = [node.clone(clone, clone) for node in self.nodes]
        return clone

    def to_block(self):
        block = Block(self, self)
        block.nodes = self.nodes
        return block

    def to_json(self):
        return json.dumps(self.__dict__)
        # return json.dumps({'__type': 'Root',
        #                    'nodes': self.nodes,
        #                    'lineno': self.lineno,
        #                    'column': self.column,
        #                    'filename': self.filename})



from stilus.nodes.color import Color
from stilus.utils import assert_color, assert_string
from stilus.nodes.unit import Unit

component_map = {
    "red": "r",
    "green": "g",
    "blue": "b",
    "alpha": "alpha",
    "hue": "hue",
    "saturation": "saturation",
    "lightness": "lightness",
}

unit_map = {"hue": "deg", "saturation": "%", "lightness": "%"}

type_map = {
    "red": "rgba",
    "green": "rgba",
    "blue": "rgba",
    "alpha": "rgba",
    "hue": "hsla",
    "saturation": "hsla",
    "lightness": "hsla",
}


def component(color: Color, name: str, evaluator=None) -> Unit:
    """Return component `name` for a given `color`.
    :param color: Color
    :param name: str
    :return: Unit
    """
    assert_color(color, "color")
    assert_string(name, "name")
    name = name.string
    unit = unit_map.setdefault(name, None)
    type = type_map[name]
    name = component_map[name]
    try:
        return Unit(getattr(getattr(color, type)(), name), unit)
    except AttributeError:
        raise TypeError(f'invalid color component "{name}"')



from typing import Type

from stilus.nodes.ident import Ident
from stilus.nodes.node import Node


class Scope:
    def __init__(self):
        self.commons = {}

    def __hash__(self):
        return hash(self.commons)

    def __eq__(self, other):
        if isinstance(other, Scope):
            return self.commons == other.commons
        return False

    def __str__(self):
        if not self.commons:
            return "[Empty scope]"
        else:
            commons = ", ".join(f"@{key}" for key in self.commons.keys())
            return f"[Scope {commons}]"

    def __repr__(self):
        return self.__str__()

    def add(self, ident: Ident):
        """Add ident node to the current scope."""
        self.commons[ident.string] = ident.value

    def lookup(self, name) -> Type[Node]:
        """Lookup the given local variable name.
        :param name:
        :return:
        """
        return self.commons.get(str(name), None)



import json

from .node import Node


class Extend(Node):
    def __init__(self, selectors, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.selectors = selectors
        self.optional = None

    def __str__(self):
        s = ", ".join(self.selectors)
        return f"@extend {s}"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.node_name, self.selectors

    def __eq__(self, other):
        if isinstance(other, Extend):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.__key())

    def clone(self, parent=None, node=Node):
        clone = Extend(self.selectors, lineno=self.lineno, column=self.column)
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Charset",
                "selectors": self.selectors,
                "prefix": self.prefix,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



import logging
import logging.config
import os
import sys
import time
from pathlib import Path
from typing import Optional, List

import click
import yaml
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from .__version__ import __version__
from .renderer import Renderer

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


def setup_logging(
    default_path="logging.yaml",
    default_level=logging.INFO,
    env_key="STILUS_LOGGING_CONFIG",
):
    """Setup logging configuration.
    :param default_path: The path to the yaml logging config file.
    :param default_level: The default level is ``logging.INFO``.  Only used
                          when the ``default_path`` is not found.
    :param env_key: The environment property to be used as default_path.
    """
    path = os.getenv(env_key, default_path)
    p = Path(path)
    if p.exists():
        with open(path) as f:
            config = yaml.safe_load(f.read())
        logging.config.dictConfig(config)
        logging.debug(f"Using {p.resolve()} as logging configuration file.")
    else:
        # todo: handle these differently to comply with Stylus
        logging.basicConfig(format="%(message)s", level=default_level)
        logging.warning(
            f"Using default logging configuration (level"
            f" is {logging.getLevelName(default_level)})."
        )
        logging.warning(
            f"Note: use STILUS_LOGGING_CONFIG to configure"
            f" the logging; or create a logging.yaml file."
        )


setup_logging("bin/logging.yaml")


def print_version(ctx, param, value):
    """
    Print the version of Stilus and end the process.

    :param ctx: The click context.
    :param param: A parameter.
    :param value: A value
    """
    if not value or ctx.resilient_parsing:
        return
    click.echo(f"{__version__}")
    ctx.exit()


def validate_path(ctx, param, value):
    return Path(value)


def fail(message, prefix=None, code=-1):
    """Show a message and end the process."""
    if not prefix:
        prefix = "     error "
    click.echo(click.style(prefix, fg="red"), nl=False)
    click.echo(message)
    sys.exit(code)


def fancy_output(message, prefix=None):
    """"""
    if prefix:
        click.echo(click.style(prefix, dim=True), nl=False)
    click.echo(message)


def render(
    source: str, includes=None, compress=False, prefix=None, ha=False
) -> str:
    """
    Render a source str to css.
    :param source: A Stylus source string.
    :type source: str
    :param includes: List of directories to include.
    :type includes: list, optional
    :param compress: Compress the output?  False by default.
    :type compress: bool, optional
    :param prefix: Prefix to add to each css class..
    :type prefix: Str, optional
    :param ha: Hoist atrules?  False by default.
    :type ha: bool, optional
    :return: The resulting css.
    :rtype: str
    """
    if not includes:
        includes = []
    renderer = Renderer(
        source, {"compress": compress, "prefix": prefix, "hoist atrules": ha}
    )
    for include in includes:
        # todo: check existence?
        renderer.include(include)
    return renderer.render()


def write_result(css, path):
    p = Path(path).resolve()
    with p.open(mode="w+"):
        p.write_text(css)


def compile(
    source: Path,
    path: Path = None,
    includes: Optional[List[Path]] = None,
    compress: bool = False,
    prefix=None,
    ha=False,
) -> None:
    """
    Compile a source Path.  If path is given compile to that path,
    otherwise compile into the same folder.
    :param source: A Stylus source Path.
    :type source: Path
    :param path: The path to compile to.
    :type path: Path, optional
    :param includes: List of directories to include.
    :type includes: list, optional
    :param compress: Compress the output?  False by default.
    :type compress: bool, optional
    :param prefix: Prefix to add to each css class..
    :type prefix: Str, optional
    :param ha: Hoist atrules?  False by default.
    :type ha: bool, optional
    """
    if path:
        destination = path / Path(source.name).with_suffix(".css")
    else:
        destination = source.with_suffix(".css")
    # todo: check includes existence here?
    css = render(source.read_text(), includes, compress, prefix, ha)
    write_result(css, destination)
    fancy_output(str(destination), prefix="  compiled ")
    logging.info(f"Compiled {destination}.")


def prepare_watch(
    path: Path,
    include: Optional[List[Path]] = None,
    compress: bool = False,
    prefix: str = None,
    ha: bool = False,
) -> List[Path]:
    """
    Compile all Stylus files in the given path and recompile
    automatically when these files change.

    :param path: A path.
    :type path: Path
    :param includes: List of directories to include.
    :type includes: list, optional
    :param compress: Compress the output?  False by default.
    :type compress: bool, optional
    :param prefix: Prefix to add to each css class.
    :type prefix: Str, optional
    :param ha: Hoist atrules?  False by default.
    :type ha: bool, optional
    :return: List of paths that are present in the given Path
    :rtype: List[Path]
    """
    if not include:
        include = []
    logging.info(f"Watching {path}...")
    styles = list(path.glob("*.styl"))
    logging.info(f"{styles}")
    for styl in styles:
        fancy_output(str(styl), prefix="  watching ")
        compile(styl, path, include, compress, prefix, ha)
    return styles


def check_out(out: str) -> None:
    """
    Check of the given out str is an existing directory.  Otherwise
    fail.
    :param out: A possibly existing directory
    :type out: str
    """
    if not Path(out).exists():
        fail(f"Directory not found: {out}.")
    elif not Path(out).is_dir():
        fail(f"{out} is not a directory.")


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("input", type=click.File("r"), required=False)
@click.argument("output", type=click.File("w"), required=False)
@click.option("-v", "--verbose", count=True, help="Be more verbose.")
@click.option(
    "-w",
    "--watch",
    is_flag=True,
    default=False,
    help="Watch file(s) for changes and re-compile.",
)
@click.option(
    "-c",
    "--compress",
    is_flag=True,
    default=False,
    help="Compress CSS output.",
)
@click.option(
    "-p",
    "--print",
    "print_",
    is_flag=True,
    default=False,
    help="Print out the compiled CSS.",
)
@click.option(
    "-I", "--include", multiple=True, help="Add <path> to lookup paths."
)
@click.option("-o", "--out", help="Output to <dir> when passing files.")
@click.option("-P", "--prefix", help="Prefix all css classes by <prefix>.")
@click.option(
    "--hoist-atrules",
    "ha",
    is_flag=True,
    default=False,
    help="Move @import and @charset to the top",
)
@click.option(
    "--version",
    "-V",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Display the version of Stilus.",
)
def stilus(
    verbose, watch, compress, print_, include, out, input, output, prefix, ha
):
    class StilusHandler(FileSystemEventHandler):
        def on_modified(self, event):
            path = Path(event.src_path)
            if path in styles:
                compile(path, None, include, compress, prefix, ha)

    if out:
        check_out(out)
        logging.info(f"Redirecting results to {out}.")

    if watch:
        path = Path.cwd()
        styles = prepare_watch(path, include, compress, prefix, ha)
        event_handler = StilusHandler()
        observer = Observer()
        observer.schedule(event_handler, path=str(path), recursive=False)
        observer.start()
        try:
            while True:
                time.sleep(0.25)
        except KeyboardInterrupt:
            logging.info("Bailing out...")
            observer.stop()
        observer.join()
    else:
        if not input:
            input = click.get_text_stream("stdin")
            logging.debug("Reading from stdin...")
        if not output:
            output = click.get_text_stream("stdout")
            logging.debug("Writing to stdout...")
        elif out:
            output = (Path(out) / Path(output.name)).open("w")
            logging.debug(f"Writing to {output}.")
        css = render(input.read(), include, compress, prefix, ha)
        if print_:
            click.echo(css, nl="\n" if compress else "")
        output.write(css)


if __name__ == "__main__":
    stilus()



import json
import re
import types
import unicodedata
from decimal import Decimal, ROUND_HALF_UP
from os.path import join
from pathlib import Path
from typing import List

from .nodes.boolean import Boolean
from .nodes.expression import Expression
from .nodes.literal import Literal
from .nodes.node import Node
from .nodes.null import null, Null
from .nodes.object_node import ObjectNode
from .nodes.string import String
from .nodes.unit import Unit
from .selector_parser import SelectorParser


def unwrap(expression: Expression) -> Node:
    """
    Unwrap `expression`.
    Takes an expressions with length of 1
    such as `((1 2 3))` and unwraps it to `(1 2 3)`.

    :param expression:
    :return:
    """
    if hasattr(expression, "preserve") and expression.preserve:
        return expression
    if hasattr(expression, "node_name") and expression.node_name not in [
        "arguments",
        "expression",
    ]:
        return expression
    if len(expression) != 1:
        return expression
    if (
        hasattr(expression, "nodes")
        and len(expression.nodes) > 0
        and expression.nodes[0].node_name not in ["arguments", "expression"]
    ):
        return expression
    if isinstance(expression, list) and len(expression) > 0 and \
            not hasattr(expression[0], 'nodes'):
        return expression
    return unwrap(expression.nodes[0])


def assert_present(node: Node, name=None):
    if node or isinstance(node, Null):
        return
    if name:
        raise ValueError(f'"{name}" argument required')
    raise ValueError("argument missing")


def assert_color(node: Node, param=Node):
    assert_present(node, param)
    if node.node_name not in ["rgba", "hsla"]:
        raise TypeError(
            f"TypeError: expected rgba or hsla, but got "
            f"{node.node_name}:{node}"
        )


def assert_string(node: Node, param):
    assert_present(node, param)
    if node.node_name in ["string", "ident", "literal"]:
        return
    else:
        raise TypeError(
            f"TypeError: expected string, ident or literal, "
            f"but got {node.node_name}:{node}"
        )


def assert_type(node: Node, type, param=None):
    assert_present(node, param)
    if node.node_name == type:
        return
    p = f"{param} to be a " if param else ""
    raise TypeError(f"expected {p}{type}, but got {node.node_name}:{node}")


def clamp(n, smallest=0, largest=255):
    return max(smallest, min(int(stilus_round(n)), largest))


def stilus_round(n):
    return Decimal(n).quantize(Decimal(1), rounding=ROUND_HALF_UP)


def clamp_alpha(n, smallest=0, largest=1):
    return max(smallest, min(n, largest))


def clamp_degrees(n):
    n = n % 360
    return n if n >= 0 else 360 + n


def clamp_percentage(n, smallest=0, largest=100):
    return max(smallest, min(n, largest))


def compile_selectors(arr, leave_hidden=False, indent=""):
    """
    Compile selector strings in `arr` from the bottom-up
    to produce the selector combinations. For example
    the following Stylus:

    '''
    ul
      li
      p
        a
          color: red
    '''

    would return:

      [ 'ul li a', 'ul p a' ]

    :param arr:
    :param leave_hidden:
    :return:
    """
    selectors = []
    indent = indent
    buf = []

    def parse(selector, buf):
        parts = [selector.value]
        parents = []
        string = SelectorParser(parts[0], parents, parts).parse()["value"]

        if buf:
            for i, part in enumerate(buf):
                parts.append(part)
                parents.append(string)
                child = SelectorParser(buf[i], parents, parts).parse()

                # todo: fix this
                if child["nested"]:
                    string += " " + child["value"]
                else:
                    string = child["value"]

        return string.strip()

    def compile(arr, i):
        if i:
            for selector in arr[i]:
                if not leave_hidden and selector.is_placeholder():
                    continue
                if selector.inherits:
                    buf.insert(0, selector.value)
                    compile(arr, i - 1)
                    buf.pop(0)
                else:
                    selectors.append(indent + parse(selector, buf))
        else:
            for selector in arr[0]:
                if not leave_hidden and selector.is_placeholder():
                    continue
                string = parse(selector, buf)
                if string:
                    selectors.append(indent + string)

    compile(arr, len(arr) - 1)

    # return unique selectors
    return list(dict.fromkeys(selectors))


def merge(a, b, deep=None):
    for k in b.keys():
        if deep and k in a.keys():
            node_a = unwrap(a[k].first())
            node_b = unwrap(b[k].first())
            if (
                node_a.node_name == "objectnode"
                and node_b.node_name == "objectnode"
            ):
                a[k].first().values = merge(node_a.values, node_b.values, deep)
            else:
                a[k] = b[k]
        else:
            a[k] = b[k]
    return a


# todo: use libpath
def lookup(path: Path, paths: List, ignore=""):
    if isinstance(path, Path):
        p = path
    else:
        p = Path(path)
    if p.is_absolute() and p.exists():
        return str(path)

    for a_path in reversed(paths):
        lookup = join(a_path, path)
        if ignore == lookup:
            continue
        p = Path(lookup).resolve()
        if p.exists():
            return str(lookup)

    return None


# fixme: this is bad
def find(path, paths, ignore):
    # first see if path exists
    p = Path(path)
    if p.is_absolute() and p.exists():
        return [str(p)]

    # if not see if it is in paths
    for base in reversed(paths):
        lookup = Path(join(base, path))
        if lookup == ignore:
            continue
        # fixme: this is *very* bad
        if "*" in path:
            found = sorted(Path(base).glob(path))
            if found:
                return [str(file) for file in found]
        else:
            if lookup.exists():
                return [str(lookup.resolve())]

    return None


def lookup_index(name, paths, filename, parser=None):
    p = Path(name)

    found = find(join(str(p), "index.styl"), paths, filename)

    if not found:
        looking_for = Path(p.name)
        looking_for = looking_for.with_suffix(".styl")
        found = find(join(str(p), str(looking_for)), paths, filename)

    def find_in_js_packages(directory):
        package = lookup(join(directory, "package.json"), paths, filename)
        if not package:
            if re.search(".styl$", directory, re.IGNORECASE):
                return lookup_index(directory, paths, filename)
            else:
                return find_in_js_packages(directory + ".styl")
        if parser:
            with open(package) as json_file:
                data = json.load(json_file)
                if data["main"]:
                    found = find(
                        join(directory, data["main"]), paths, filename
                    )
                else:
                    found = lookup_index(directory, paths, filename)
        return found

    if not found and "node_modules" not in name:
        found = find_in_js_packages(join("node_modules", name))

    return found


def coerce_list(value, raw, lineno=1, column=1):
    expr = Expression(lineno=lineno, column=column)
    for v in value:
        expr.append(coerce(v, raw, lineno=lineno, column=column))
    return expr


def coerce_object(object, raw, lineno=1, column=1):
    # todo: this needs tests!
    node = ObjectNode() if raw else Expression()
    for key in object:
        value = coerce(object[key], raw, lineno=lineno, column=column)
        if raw:
            node[key] = value
        else:
            node.append(coerce_list([key, value], False))
    return node


def coerce(value, raw: bool, lineno=1, column=1):
    if isinstance(value, types.FunctionType):
        return value
    elif isinstance(value, str):
        return String(value, lineno=lineno, column=column)
    elif isinstance(value, bool):
        return Boolean(value, lineno=lineno, column=column)
    elif isinstance(value, int):
        return Unit(value, lineno=lineno, column=column)
    elif isinstance(value, list):
        return coerce_list(value, raw, lineno=lineno, column=column)
    elif value is None:
        return null
    elif hasattr(value, "node_name"):
        return value
    else:
        coerce_object(value, raw, lineno=lineno, column=column)


def parse_string(str: String):
    from stilus.parser import Parser

    try:
        parser = Parser(str, {})
        result = parser.list()
    except BaseException:
        result = Literal(str)
    return result


def get_value(node: Node):
    try:
        f = float(node.value)
        v = int(f) if f.is_integer() else f
    except ValueError:
        v = node.value()
    return v


def is_number(node: Node):
    try:
        float(node.value)
        return True
    except (TypeError, ValueError):
        pass
    try:
        unicodedata.numeric(node.value)
        return True
    except (TypeError, ValueError):
        pass
    return False



import re
from os.path import abspath, dirname

from stilus import utils
from stilus.exceptions import StilusError
from stilus.nodes.block import Block
from stilus.nodes.boolean import Boolean
from stilus.nodes.call import Call
from stilus.nodes.expression import Expression
from stilus.nodes.group import Group
from stilus.nodes.ident import Ident
from stilus.nodes.media import Media
from stilus.nodes.node import Node
from stilus.nodes.property import Property
from stilus.nodes.root import Root
from stilus.nodes.string import String
from stilus.nodes.unit import Unit, has_many_zeros_and_a_number
from stilus.visitor.visitor import Visitor


# fixme: this method should not be called
def _handle_weird_deg_value(value, type):
    """
    Remove all numbers after the period (and the period itself) from
    deg units that have 14 zeros after the period and one single
    non-zero character at the end.

    This is a hack; it is caused by the colorsys call in the Color
    node.
    """
    result = value
    if type == "deg" and re.match(r"[1-9]", value[-1]):
        match = re.search(r"\.(0*)", value)
        if match and len(match.groups()) == 1 and len(match.group(1)) == 14:
            result = value[:-1].rstrip("0").rstrip(".")
    return result


# fixme: this method should not exist!
def _handle_weird_unit(v, n, unit):
    if (
        isinstance(n, float)
        and unit.value is not None
        and has_many_zeros_and_a_number(v)
    ):
        return unit.value
    return v


class Compiler(Visitor):
    def __init__(self, root, options):
        super().__init__(root)
        self.compress = options.get("compress", False)
        self.linenos = options.get("linenos", False)
        self.spaces = options.get("indent spaces", 2)
        self.indents = 0
        self.buf = None
        self.last = None
        self.keyframe = None
        self.is_condition = None
        self.is_url = None
        self.stack = []

    def compile(self):
        return self.visit(self.root)

    def out(self, string, node=None):
        return string

    def indent(self):
        if self.compress:
            return ""
        return " " * (self.spaces * self.indents)

    def need_brackets(self, node):
        if 0 == self.indents:
            return True
        if "atrule" != node.node_name:
            return True
        if node.has_only_properties():
            return True
        return False

    def visit_root(self, root: Root):
        self.buf = ""
        for node in root.nodes:
            if self.linenos:
                self.debug_info(node)
            ret = self.visit(node)
            if ret:
                self.buf += self.out(ret + "\n", node)
        return self.buf

    def visit_block(self, block: Block):  # noqa: C901
        if self.compress:
            separator = ""
        else:
            separator = "\n"

        if block.has_properties() and not block.lacks_rendered_selectors:

            last_property_index = 0
            if self.compress:
                for i, node in reversed(list(enumerate(block.nodes))):
                    if node.node_name == "property":
                        last_property_index = i
                        break

            needs_brackets = self.need_brackets(block.node)
            if needs_brackets:
                self.buf += self.out("{" if self.compress else " {\n")
                self.indents += 1

            for i, node in enumerate(block.nodes):
                self.last = last_property_index == i
                if node.node_name in [
                    "null",
                    "expression",
                    "function",
                    "group",
                    "block",
                    "unit",
                    "media",
                    "keyframes",
                    "atrule",
                    "supports",
                ]:
                    continue
                elif (
                    node.node_name == "comment"
                    and node.inline
                    and not self.compress
                ):
                    self.buf = self.buf[0:-1]
                    self.buf += self.out(f" {self.visit(node)}\n", node)
                elif node.node_name == "property":
                    ret = self.visit(node) + separator
                    self.buf += ret if self.compress else self.out(ret, node)
                else:
                    self.buf += self.out(self.visit(node) + separator, node)

            if needs_brackets:
                self.indents -= 1
                self.buf += self.out(self.indent() + "}" + separator)

        # nesting
        index = 0
        while index < len(block.nodes):
            node = block.nodes[index]
            if node.node_name in ["group", "block", "keyframes"]:
                if self.linenos:
                    self.debug_info(node)
                self.visit(node)
            elif node.node_name in ["media", "import", "atrule", "supports"]:
                self.visit(node)
            elif node.node_name == "comment":
                if not node.suppress:
                    self.buf += self.out(
                        self.indent() + self.visit(node) + "\n", node
                    )
            elif node.node_name in ["charset", "literal", "namespace"]:
                self.buf += self.out(self.visit(node) + "\n", node)
            index += 1

    def visit_keyframes(self, node: Node):
        if not node.frames:
            return

        if node.prefix == "official":
            prefix = ""
        else:
            prefix = f"-{node.prefix}-"

        brace = "{" if self.compress else " {\n"
        self.buf += self.out(
            f"@{prefix}keyframes " f"{self.visit(node.value)}{brace}", node
        )

        self.keyframe = True
        self.indents += 1
        self.visit(node.block)
        self.indents -= 1
        self.keyframe = False

        brace = "" if self.compress else "\n"
        self.buf += self.out(f"}}{brace}")

    def visit_media(self, media: Media):
        val = media.value
        if not media.has_output() or len(val.nodes) == 0:
            return

        self.buf += self.out("@media ", media)
        self.visit(val)
        self.buf += self.out("{" if self.compress else " {\n")
        self.indents += 1
        self.visit(media.block)
        self.indents -= 1
        self.buf += self.out("}" if self.compress else "}\n")

    def visit_querylist(self, queries):
        for i, node in enumerate(queries.nodes):
            self.visit(node)
            if len(queries.nodes) - 1 != i:
                self.buf += self.out("," if self.compress else ", ")

    def visit_query(self, node):
        length = len(node.nodes)
        if node.predicate:
            self.buf += self.out(f"{node.predicate} ")
        if node.type:
            self.buf += self.out(
                str(node.type) + (" and " if length > 0 else "")
            )
        for i, node in enumerate(node.nodes):
            self.buf += self.out(self.visit(node))
            if length - 1 != i:
                self.buf += self.out(" and ")

    def visit_feature(self, node: Node):
        if node.expr is None:
            return node.name
        elif node.expr.is_empty():
            return "(" + node.name + ")"
        else:
            return (
                "("
                + node.name
                + (":" if self.compress else ": ")
                + self.visit(node.expr)
                + ")"
            )

    def visit_import(self, imported):
        self.buf += self.out(
            "@import " + self.visit(imported.path) + ";\n", imported
        )

    def visit_atrule(self, atrule):
        newline = "" if self.compress else "\n"

        self.buf += self.out(f"{self.indent()}@{atrule.type}", atrule)

        if atrule.value:
            self.buf += self.out(f" {atrule.value.strip()}")

        if atrule.block:
            if atrule.has_only_properties():
                self.visit(atrule.block)
            else:
                self.buf += self.out("{" if self.compress else " {\n")
                self.indents += 1
                self.visit(atrule.block)
                self.indents -= 1
                self.buf += self.out(self.indent() + "}" + newline)
        else:
            self.buf += self.out(";" + newline)

    def visit_supports(self, node):
        if not node.has_output():
            return

        self.buf += self.out(self.indent() + "@supports ", node)
        self.is_condition = True
        self.buf += self.out(self.visit(node.condition))
        self.is_condition = False
        self.buf += self.out("{" if self.compress else " {\n")
        self.indents += 1
        self.visit(node.block)
        self.indents -= 1
        self.buf += self.out(self.indent() + ("}" if self.compress else "}\n"))

    def visit_comment(self, comment):
        if self.compress:
            if comment.suppress:
                return ""
            else:
                return comment.value
        else:
            return comment.value

    def visit_function(self, fn):
        return fn.function_name

    def visit_charset(self, charset):
        return "@charset " + self.visit(charset.value) + ";"

    def visit_namespace(self, namespace):
        prefix = self.visit(namespace.prefix) + " " if namespace.prefix else ""
        return f"@namespace {prefix}{self.visit(namespace.value)};"

    def visit_literal(self, literal):
        val = literal.value
        if literal.css:
            val = re.sub(r"^  ", "", val, flags=re.MULTILINE)
        return val

    def visit_boolean(self, boolean: Boolean):
        return str(boolean)

    def visit_rgba(self, rgba):
        return str(rgba)

    def visit_hsla(self, hsla):
        return str(hsla.rgba())

    def visit_unit(self, unit: Unit):
        t = unit.type if unit.type else ""
        n = unit.value
        f = isinstance(n, float)
        if self.compress:
            # always return '0' unless the unit is a percentage or time
            if "%" != t and "s" != t and "ms" != t and 0 == n:
                return 0
            if f:
                # remove trailing zeros
                n = f"{n:.3f}".rstrip("0").rstrip(".")
                if 1 > float(n) > -1:
                    # omit leading '0' on floats
                    n = n.replace("0.", ".")
                return f"{n}{t}"
        if f:
            v = f"{n:.15f}".rstrip("0").rstrip(".")
            v = _handle_weird_deg_value(v, t)
            v = _handle_weird_unit(v, n, unit)  # fixme: remove this!
            return f"{v}{t}"
        else:
            return f"{n}{t}"

    def visit_group(self, group: Group):
        if self.keyframe:
            stack = []
        else:
            stack = self.stack

        if self.compress:
            comma = ","
        else:
            comma = ",\n"

        stack.append(group.nodes)

        # selectors
        if group.block.has_properties():
            selectors = utils.compile_selectors(
                stack, leave_hidden=False, indent=self.indent()
            )

            if selectors:
                if self.keyframe:
                    if self.compress:
                        comma = ","
                    else:
                        comma = ", "

                for i, selector in enumerate(selectors):
                    last = i == len(selectors) - 1

                    # keyframe bocks (10%, 20% { ... })
                    if self.keyframe:
                        selector = selector.strip() if i else selector
                    try:
                        self.buf += self.out(
                            selector + ("" if last else comma)
                        )
                    except BaseException:
                        raise StilusError()
            else:
                group.block.lacks_rendered_selectors = True

        self.visit(group.block)
        stack.pop()

    def visit_ident(self, ident: Ident):
        return ident.name

    def visit_string(self, string: String):
        return string.value if self.is_url else str(string)

    def visit_null(self, node: Node):
        return ""

    def visit_call(self, call: Call):
        self.is_url = "url" == call.function_name
        args = [self.visit(node) for node in call.args.nodes]
        if self.compress:
            args = ",".join(args)
        else:
            args = ", ".join(args)
        if self.is_url:
            args = f'"{args}"'
        self.is_url = False
        return f"{call.function_name}({args})"

    def visit_expression(self, expr: Expression):
        buf = []
        length = len(expr.nodes)
        nodes = [self.visit(node) for node in expr.nodes]

        for i, (node, next) in enumerate(zip(nodes, nodes[1:] + ["end"])):
            last = i == length - 1
            buf.append(str(node))
            if "/" == next or "/" == node:
                continue
            if last:
                continue

            if self.is_url or (
                self.is_condition and (")" == next or "(" == node)
            ):
                space = ""
            else:
                space = " "

            if expr.is_list:
                if self.compress:
                    buf.append(",")
                else:
                    buf.append(", ")
            else:
                buf.append(space)

        return "".join(buf)

    def visit_arguments(self, arguments):
        return self.visit_expression(arguments)

    def visit_property(self, property: Property):
        val = self.visit(property.expr).strip()
        if property.name:
            name = property.name
        else:
            s = [f"{segment}" for segment in property.segments]
            name = "".join(s)
        arr = []
        arr.append(self.out(self.indent()))
        arr.append(self.out(name + (":" if self.compress else ": ")))
        arr.append(self.out(val, property.expr))
        if self.last:
            separator = ""
        else:
            separator = ";"
        arr.append(self.out(separator if self.compress else ";"))
        return "".join(arr)

    def debug_info(self, node: Node):
        if node.filename == "stdin":
            path = "stdin"
        else:
            path = dirname(abspath(node.filename))
        line = node.nodes[0].lineno if node.nodes else node.lineno
        if self.linenos:
            self.buf += f"\n/* line {line} : {path} */\n"



from stilus.functions.component import component
from stilus.functions.hsla import hsla
from stilus.nodes.color import Color
from stilus.nodes.string import String
from stilus.nodes.unit import Unit


def lightness(color: Color, value=None, evaluator=None):
    if value:
        hsla_color = color.hsla()
        return hsla(
            Unit(hsla_color.hue),
            Unit(hsla_color.saturation),
            value,
            Unit(hsla_color.a),
            evaluator,
        )
    return component(color, String("lightness"), evaluator)



from stilus.nodes.color import HSLA
from stilus.utils import assert_color, assert_type


def hsla(hue, saturation=None, lightness=None, alpha=None, evaluator=None):
    """Convert the given `color` to an `HSLA` node, or h,s,l,a component
    values.
    :param hue:
    :param saturation:
    :param lightness:
    :param alpha:
    :return: :class:HSLA
    """
    # color
    if hue and (saturation is lightness is alpha is None):
        assert_color(hue)
        return hue.hsla()
    # alpha
    if hue and saturation and (lightness is alpha is None):
        assert_color(hue)
        color = hue.hsla()
        assert_type(saturation, "unit", "alpha")
        alpha = saturation.clone()
        if alpha.type == "%":
            alpha.value /= 100
        return HSLA(color.hue, color.saturation, color.lightness, alpha.value)
    # color
    assert_type(hue, "unit", "hue")
    assert_type(saturation, "unit", "saturation")
    assert_type(lightness, "unit", "lightness")
    assert_type(alpha, "unit", "alpha")
    alpha = alpha.clone()
    if alpha and alpha.type == "%":
        alpha.value /= 100
    return HSLA(hue.value, saturation.value, lightness.value, alpha.value)



from stilus import utils
from stilus.exceptions import StilusError
from stilus.nodes.literal import Literal


def opposite_position(positions, evaluator=None):
    expr = []
    nodes = utils.unwrap(positions)
    for i, node in enumerate(nodes):
        utils.assert_string(node, f"position {i}")
        if node.string == "top":
            expr.append(Literal("bottom"))
        elif node.string == "bottom":
            expr.append(Literal("top"))
        elif node.string == "left":
            expr.append(Literal("right"))
        elif node.string == "right":
            expr.append(Literal("left"))
        elif node.string == "center":
            expr.append(Literal("center"))
        else:
            raise StilusError(f"invalid position {i}")
    return expr



import json

from stilus import utils
from .boolean import Boolean
from .expression import Expression
from .node import Node


class String(Node):
    def __init__(self, value, quote=None, lineno=1, column=1):
        super().__init__(value, lineno=lineno, column=column)
        self.string = value
        self.prefixed = False
        if not isinstance(quote, str):
            self.quote = "'"
        else:
            self.quote = quote

    def __str__(self):
        return f"{self.quote}{self.value}{self.quote}"

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(str(self))

    def clone(self, parent=None, node=None):
        clone = String(
            self.value, self.quote, lineno=self.lineno, column=self.column
        )
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Boolean",
                "value": self.value,
                "quote": self.quote,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )

    def to_boolean(self):
        return Boolean(len(self.value))

    def coerce(self, other: Node):
        if other.node_name == "string":
            return other
        elif other.node_name == "expression":
            node_values = [self.coerce(node).value for node in other.nodes]
            return String(" ".join(node_values))
        else:
            return String(other.__str__())

    def operate(self, op, right, value=None):
        if op == "%":
            expr = Expression()
            expr.append(self)
            # constructor args
            if right.node_name == "expression":
                args = utils.unwrap(right)
            else:
                args = [right]
            # apply
            from stilus.functions.s import s

            if hasattr(args, "nodes"):
                return s(expr, *args.nodes)
            else:
                return s(expr, args[0])
        if op == "+":
            expr = Expression()
            expr.append(
                String(
                    self.value + self.coerce(right).value,
                    lineno=self.lineno,
                    column=self.column,
                )
            )
            return expr
        else:
            return super().operate(op, right)



from typing import Type

from stilus.nodes.block import Block
from stilus.nodes.node import Node
from stilus.stack.frame import Frame


class Stack(list):
    """A stack of Frames."""

    def __init__(self):
        """Create an empty ``Frame`` stack"""
        super().__init__()

    def __str__(self):
        buffer = []
        for frame in self:
            node = frame.block.node
            location = f"({node.filename}:{node.lineno}:{node.column})"
            if node.name == "function":
                buffer.append(f"    at {node.node_name}() {location}")
            elif node.name == "group":
                buffer.append(f'    at "{node.nodes[0].value}" {location}')
        return "\n".join(buffer)

    def __repr__(self):
        return self.__str__()

    def __next__(self):
        return super(next())

    def append(self, frame: Frame):
        """Push a frame to the stack."""
        frame.stack = self
        frame.parent = self.current_frame()
        return super().append(frame)

    def current_frame(self: Frame):
        """Get the current frame."""
        if len(self) == 0:
            return None
        else:
            return self[-1]

    def get_block_frame(self, block: Block):
        """Lookup stack frame for a given block"""
        for frame in self:
            if block == frame.block:
                return frame
        return None

    def lookup(self, name) -> Type[Node]:
        """Lookup the given local variable `name`, relative to the lexical
        scope of the current frame's `Block`.
        When the result of a lookup is an identifier a recursive lookup
        is performed, defaulting to returning the identifier itself.
        :param name: local variable
        :return:
        """
        block = None
        if self.current_frame():
            block = self.current_frame().block
        while True:
            frame = self.get_block_frame(block)
            if frame:
                val = frame.lookup(name)
                if val is not None:  # fixme: use val: (fix __len__)
                    return val
            if hasattr(block, "parent"):
                block = block.parent
            else:
                block = None
            if block is None:
                return



import colorsys
import json

from stilus.functions.adjust import adjust
from stilus.utils import clamp, clamp_alpha, clamp_degrees, clamp_percentage
from .boolean import Boolean, true
from .node import Node
from .string import String
from .unit import Unit


class Color(Node):
    def __init__(self, a):
        super().__init__()
        self.a = clamp_alpha(a)

    def rgba(self):
        pass


class HSLA(Color):
    def __init__(self, h, s, l, a):
        super().__init__(a)
        self.hue = clamp_degrees(h)
        self.saturation = clamp_percentage(s)
        self.lightness = clamp_percentage(l)
        self._hsla = self

    def hsla(self):
        return self._hsla

    def _cleanup(self, value):
        if int(value) == value:
            return int(value)
        else:
            return value

    def __str__(self):
        return (
            f"hsla({self._cleanup(self.hue)},"
            f"{self._cleanup(round(self.saturation))}%,"
            f"{self._cleanup(round(self.lightness))}%,"
            f"{self._cleanup(self.a)})"
        )

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if isinstance(other, HSLA):
            return (
                self.hue == other.hue
                and self.saturation == other.saturation
                and self.lightness == other.lightness
                and self.a == other.a
            )
        return False

    def clone(self, parent=None, node=Node):
        clone = HSLA(self.hue, self.saturation, self.lightness, self.a)
        clone.lineno = self.lineno
        clone.column = self.column
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "HSLA",
                "h": self.hue,
                "s": self.saturation,
                "l": self.lightness,
                "a": self.a,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )

    def rgba(self):
        return RGBA.from_hsla(self._hsla)

    def __hash__(self):
        return hash(self.__str__())

    def hash(self):
        return self.rgba().hash()

    def add(self, h, s, l):
        return HSLA(
            self.hue + h, self.saturation + s, self.lightness + l, self.a
        )

    def sub(self, h, s, l):
        return self.add(-h, -s, -l)

    def operate(self, op, right, value=None):
        """Operate on `right` with the given `op`."""
        if op in ["==", "!=", "<=", ">=", "<", ">", "is a", "||", "&&"]:
            return self.rgba().operate(op, right)
        else:
            return self.rgba().operate(op, right).hsla()

    @staticmethod
    def from_rgba(rgba):
        (h, l, s) = colorsys.rgb_to_hls(
            rgba.r / 255, rgba.g / 255, rgba.b / 255
        )
        return HSLA(h * 360, s * 100, l * 100, rgba.a)

    def adjust_lightness(self, percent):
        self.lightness = clamp_percentage(
            self.lightness + self.lightness * (percent / 100)
        )
        return self

    def adjust_hue(self, degree):
        self.hue = clamp_degrees(self.hue + degree)
        return self


class RGBA(Color):
    def __init__(self, r, g, b, a):
        super().__init__(a)
        self.r = clamp(r)
        self.g = clamp(g)
        self.b = clamp(b)
        self.name = ""
        self._rgba = self
        self.raw = None

    def __str__(self):
        def pad(n):
            return f"{n:02x}"

        if self.name == "transparent":
            return self.name

        if 1 == self.a:
            r = pad(self.r)
            g = pad(self.g)
            b = pad(self.b)

            if r[0] == r[1] and g[0] == g[1] and b[0] == b[1]:
                return f"#{r[0]}{g[0]}{b[0]}"
            else:
                return f"#{r}{g}{b}"
        else:
            alpha = f"{self.a:.3f}".rstrip("0").rstrip(".")
            return f"rgba({self.r},{self.g},{self.b},{alpha})"

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def without_clamping(r, g, b, a):
        rgba = RGBA(0, 0, 0, 0)
        rgba.r = r
        rgba.g = g
        rgba.b = b
        rgba.a = a
        return rgba

    def clone(self, parent=None, node=None):
        clone = RGBA(self.r, self.g, self.b, self.a)
        clone.raw = self.raw
        clone.name = self.name
        clone.lineno = self.lineno
        clone.column = self.column
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "RGBA",
                "r": self.r,
                "g": self.g,
                "b": self.b,
                "a": self.a,
                "raw": self.raw,
                "node_name": self.name,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )

    def to_boolean(self):
        return true

    def __hash__(self):
        return hash(str(self))

    def hash(self):
        return str(self)

    def hsla(self):
        return HSLA.from_rgba(self.rgba())

    def rgba(self):
        return self._rgba

    def add(self, r, g, b, a):
        return RGBA(self.r + r, self.g + g, self.b + b, self.a + a)

    def sub(self, r, g, b, a):
        return RGBA(
            self.r - r,
            self.g - g,
            self.b - b,
            self.a if a == 1 else self.a - a,
        )

    def multiply(self, n):
        return RGBA(self.r * n, self.g * n, self.b * n, self.a)

    def divide(self, n):
        return RGBA(self.r / n, self.g / n, self.b / n, self.a)

    def __eq__(self, other):
        if isinstance(other, RGBA):
            return (
                self.r == other.r and self.g == other.g and self.b == other.b
            )
        return False

    @staticmethod
    def from_hsla(hsla: HSLA):
        (r, g, b) = colorsys.hls_to_rgb(
            hsla.hue / 360, hsla.lightness / 100, hsla.saturation / 100
        )
        return RGBA(r * 255, g * 255, b * 255, hsla.a)

    def operate(self, op, right: Node, value=None):  # noqa: C901
        """Operate on `right` with given `op`."""
        if op != "in":
            right = right.first()
        if (
            op == "is a"
            and right.node_name == "string"
            and right.string == "color"
        ):
            return Boolean(true)
        elif op == "+":
            if right.node_name == "unit":
                n = right.value
                if right.type == "%":
                    return adjust(self, String("lightness"), right)
                elif right.type == "deg":
                    return self.hsla().adjust_hue(n).rgba()
                else:
                    return self.add(n, n, n, 0)
            elif right.node_name == "rgba":
                return self.add(right.r, right.g, right.b, right.a)
            elif right.node_name == "hsla":
                return self.hsla().add(
                    right.hue, right.saturation, right.lightness
                )
        elif op == "-":
            if right.node_name == "unit":
                n = right.value
                if right.type == "%":
                    return adjust(self, String("lightness"), Unit(-n, "%"))
                elif right.type == "deg":
                    return self.hsla().adjust_hue(-n).rgba()
                else:
                    return self.sub(n, n, n, 0)
            elif right.node_name == "rgba":
                return self.sub(right.r, right.g, right.b, right.a)
            elif right.node_name == "hsla":
                return self.hsla().sub(
                    right.hue, right.saturation, right.lightness
                )
        elif op == "*":
            if right.node_name == "unit":
                return self.multiply(right.value)
        elif op == "/":
            if right.node_name == "unit":
                return self.divide(right.value)
        return super().operate(op, right)



import json

from stilus.functions.bifs import bifs
from .node import Node


class Function(Node):
    def __init__(
        self,
        function_name,
        params=None,
        body=None,
        builtin=None,
        lineno=1,
        column=1,
    ):
        super().__init__(lineno=lineno, column=column)
        self.function_name = function_name
        self.params = params
        if builtin is not None:
            self.builtin = builtin
        else:
            self.builtin = self.function_name in bifs.keys()
        self.block = body
        self.fn = None
        if hasattr(params, "__call___"):
            self.fn = params

    def __str__(self):
        if self.params:
            if hasattr(self.params, "nodes"):
                strings = [
                    str(node if node else "None") for node in self.params.nodes
                ]
                return f'{self.function_name}({", ".join(strings)})'
            else:
                return f"{self.function_name}()"
        else:
            return f"{self.function_name}()"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.node_name, self.function_name, self.params, self.block

    def __eq__(self, other):
        if isinstance(other, Function):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.__key())

    def hash(self):
        return f"function {self.function_name}"

    def clone(self, parent=None, node=None):
        if self.fn:
            # check this: what is fn?
            clone = Function(self.function_name, self.fn)
        else:
            clone = Function(self.function_name)
            if hasattr(self.params, "clone"):
                clone.params = self.params.clone(parent, clone)
            if hasattr(self.block, "clone"):
                clone.block = self.block.clone(parent, clone)
        clone.lineno = self.lineno
        clone.column = self.column
        clone.filename = self.filename
        return clone

    def arity(self):
        return len(self.params)

    def to_json(self):
        return json.dumps(
            {
                "__type": "Function",
                "node_name": self.function_name,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
                "params": self.params,
                "block": self.block,
            }
        )



from stilus import utils
from stilus.utils import assert_present, assert_type, unwrap


def merge(dest, *args, evaluator=None):
    assert_present(dest, "dest")
    dest = unwrap(dest).first()
    assert_type(dest, "objectnode", "dest")

    last = unwrap(args[len(args) - 1]).first()
    deep = True is last.value

    for arg in args:
        values = unwrap(arg).first()
        if hasattr(values, "values"):
            utils.merge(dest.values, unwrap(arg).first().values, deep)

    return dest



from stilus.nodes.unit import Unit
from stilus.utils import assert_string


def unit(unit, type=None, evaluator=None):
    # assert_type(unit, 'unit', 'unit')

    if type:
        assert_string(type, "type")
        return Unit(unit.value, type.string)
    else:
        return unit.type if hasattr(unit, "type") and unit.type else ""



from stilus.functions.rgba import rgba
from stilus.nodes.color import Color, HSLA
from stilus.nodes.unit import Unit


def blue(color: Color, value: Unit = None, evaluator=None):
    """Return the blue component of the given `color`, or set the blue
    component to the optional second `value` argument.

    >>> blue(#00c)
    204

    >>> blue(#000, 255)
    #00f

    :param color:
    :param value:
    :return:
    """
    if isinstance(color, HSLA):
        color = color.rgba()
    if value:
        return rgba(
            Unit(color.r), Unit(color.g), value, Unit(color.a), evaluator
        )
    return Unit(color.b, "")



import json

from .atrule import Atrule


class Media(Atrule):
    def __init__(self, value=None, lineno=1, column=1):
        super().__init__("media", lineno=lineno, column=column)
        self.value = value

    def __str__(self):
        return f"@media {self.value}"

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, Media):
            return self.__key() == other.__key()
        return False

    def clone(self, parent=None, noe=None):
        clone = Media(lineno=self.lineno, column=self.column)
        clone.value = self.value.clone(parent, clone)
        clone.block = self.block.clone(parent, clone)
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Media",
                "val": self.value,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



from stilus.nodes.expression import Expression
from stilus.nodes.literal import Literal
from stilus.utils import unwrap, assert_type


def _convert(num, base):
    convert_string = "0123456789abcdefghijklmnopqrstuvwxyz"
    if num < base:
        return convert_string[num]
    else:
        return _convert(num // base, base) + convert_string[num % base]


def base_convert(num, base, width=None, evaluator=None):
    if isinstance(num, Expression):
        num = unwrap(num).nodes[0]
    if isinstance(base, Expression):
        base = unwrap(base).nodes[0]
    if width and isinstance(width, Expression):
        width = unwrap(width).nodes[0]
    assert_type(num, "unit")
    assert_type(base, "unit")
    if width:
        assert_type(width, "unit")
    if width:
        width = width.value
    else:
        width = 2
    num = int(num.value)
    base = int(base.value)
    result = _convert(num, base)
    while len(result) < width:
        result = "0" + result
    return Literal(result)



import json

from .boolean import Boolean, false
from .node import Node
from .null import null


class ObjectNode(Node):
    def __init__(self, values: dict = None, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        if values is None:
            self.values = {}
        else:
            self.values = values

    def __str__(self):
        s = []
        for key, value in self.values.items():
            s.append(f'"{key}":"{str(self.values[key])}"')
        return "{" + ",".join(s) + "}"

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return self.values

    def __eq__(self, other):
        if isinstance(other, ObjectNode):
            return self.__key() == other.__key()
        return False

    def set(self, key, val):
        self.values[key] = val
        return self

    def get(self, key):
        return self.values.get(key, null)

    def has(self, key):
        return key in self.values

    def __len__(self):
        return len(self.values)

    def __contains__(self, key):
        return key in self.values

    def operate(self, op, right):
        if op in [".", "[]"]:
            return self.get(right.hash())
        elif op == "==":
            values = self.values
            if right.node_name != "objectnode" or len(self) != len(right):
                return false
            for key in values:
                a = values[key]
                b = right.values[key]
                if a.operate(op, b).is_false():
                    return false
            return Boolean(True)
        elif op == "!=":
            return self.operate("==", right).negate()
        else:
            return self.operate(op, right)

    def to_boolean(self):
        return Boolean(len(self))

    def to_block(self):

        # fixme: node.nodes
        def to_string(node):
            if hasattr(node, "nodes") and node.nodes:
                strings = []
                for n in node.nodes:
                    strings.append(to_string(n))
                if hasattr(node, "is_list") and node.is_list:
                    return ",".join(strings)
                else:
                    return " ".join(strings)
            elif node.node_name == "literal" and node.value == ",":
                return "\\,"
            else:
                return str(node)

        string = "{"
        for key, value in self.values.items():
            if value.first().node_name == "objectnode":
                string += key + " " + value.first().to_block()
            else:
                if key == "@charset":
                    string += key + " " + str(value.first) + ";"
                else:

                    string += key + ":" + to_string(value) + ";"
        string += "}"
        return string

    def clone(self, parent=None, node=None):
        clone = ObjectNode(lineno=self.lineno, column=self.column)
        clone.filename = self.filename
        clone.values = {
            key: value.clone(parent, clone)
            for key, value in self.values.items()
        }
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Object",
                "vals": self.values,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



import json

from .node import Node


class Block(Node):
    def __init__(self, parent, node, scope=True, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.nodes = []
        self.parent = parent
        self.node = node
        self.scope = scope
        self.lacks_rendered_selectors = False  # for compiler
        self.index = 0
        self.mixin = False

    def __str__(self):
        return (
            f"block [{self.lineno}:{self.column}] | scope: {self.scope}; "
            f"node: {self.node.node_name}"
        )

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.nodes, self.node, self.parent, self.scope

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, Block):
            return id(self) == id(other)
        return False

    def has_properties(self):
        for node in self.nodes:
            if node.node_name == "property":
                return True
        return False

    def has_media(self):
        for node in self.nodes:
            if node.node_name == "media":
                return True
        return False

    def __len__(self):
        return len(self.nodes)

    def is_empty(self):
        return len(self.nodes) == 0

    def clone(self, parent=None, node=None):
        p = parent
        n = node
        if p is None:
            p = self.parent
        if n is None:
            n = self.node
        clone = Block(p, n, lineno=self.lineno, column=self.column)
        clone.filename = self.filename
        clone.scope = self.scope
        for something in self.nodes:
            clone.nodes.append(something.clone(clone, clone))
        return clone

    def append(self, node):
        self.nodes.append(node)

    def to_json(self):
        return json.dumps(
            {
                "__type": "Block",
                "scope": self.scope,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
                "nodes": self.nodes,
            }
        )



import json

from .node import Node


class Params(Node):
    def __init__(self, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.nodes = []

    def __str__(self):
        return f"nothing"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.node_name

    def __eq__(self, other):
        if isinstance(other, Params):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.__key())

    def __len__(self):
        return len(self.nodes)

    def append(self, node):
        self.nodes.append(node)

    def clone(self, parent=None, node=Node):
        clone = Params(lineno=self.lineno, column=self.column)
        clone.filename = self.filename
        for something in self.nodes:
            clone.nodes.append(something.clone(parent, clone))
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Params",
                "nodes": self.nodes,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



from stilus.nodes.color import RGBA
from stilus.utils import assert_color, assert_type


def rgba(red, green=None, blue=None, alpha=None, evaluator=None):
    """Return a `RGBA` from the r,g,b,a channels.

    >>> rgba(255,0,0,0.5)
    rgba(255,0,0,0.5)
    >>> rgba(255,0,0,1)
    #ff0000
    >>> rgba(#ffcc00, 50%)
    rgba(255,204,0,0.5)

    :param red:
    :param green:
    :param blue:
    :param alpha:
    :return:
    """
    # color
    if red and (green is blue is alpha is None):
        assert_color(red)
        return red.rgba()
    # alpha
    if red and green and (blue is alpha is None):
        assert_color(red)
        color = red.rgba()
        assert_type(green, "unit", "alpha")
        alpha = green.clone()
        if alpha.type == "%":
            alpha.value /= 100
        return RGBA(color.r, color.g, color.b, alpha.value)
    # color
    assert_type(red, "unit", "red")
    assert_type(green, "unit", "green")
    assert_type(blue, "unit", "blue")
    assert_type(alpha, "unit", "alpha")
    r = round(red.value * 2.55) if red.type == "%" else red.value
    g = round(green.value * 2.55) if green.type == "%" else green.value
    b = round(blue.value * 2.55) if blue.type == "%" else blue.value

    alpha = alpha.clone()
    if alpha and alpha.type == "%":
        alpha.value /= 100

    return RGBA(r, g, b, alpha.value)



from stilus.nodes.null import null
from stilus.utils import assert_type


def lookup(name, evaluator=None):
    assert_type(name, "string", "name")
    node = evaluator.lookup(name.value)
    if node:
        return evaluator.visit(node)
    else:
        return null



from deprecated import deprecated

from stilus.nodes.block import Block
from stilus.stack.scope import Scope


# todo: create getter and setter for scope or at least clean this up!
class Frame:
    def __init__(self, block: Block, parent: Block = None):
        if hasattr(block, "scope") and block.scope is False:
            self._scope = None
        else:
            self._scope = Scope()
        self.block = block
        self.parent = parent

    def __key(self):
        return self._scope, self.block, self.parent

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, Frame):
            return self.__key() == other.__key()
        return False

    def __str__(self):
        if hasattr(self.block, "scope") and self.block.scope is False:
            scope = "scope-less"
        else:
            scope = f"{self.scope()}"
        return f"[Frame {scope}]"

    def __repr__(self):
        return self.__str__()

    def scope(self):
        """Return this frame's scope or the parent scope
        for scope-less blocks.
        :return:
        """
        if self._scope:
            return self._scope
        try:
            return self.parent.scope()
        except AttributeError:
            raise TypeError("Cannot read property 'scope' of undefined")

    @deprecated
    def set_scope(self, scope: Scope):
        self._scope = scope

    def lookup(self, name):
        """
        Lookup the given local variable `name`.
        :param name:
        :return:
        """
        return self.scope().lookup(name)



import inspect
import logging
import os
import re
from pathlib import Path

from stilus import utils
from stilus.colors import colors
from stilus.exceptions import ParseError, StilusError
from stilus.functions.bifs import bifs, raw_bifs
from stilus.nodes import node as filename_node
from stilus.nodes.arguments import Arguments
from stilus.nodes.block import Block
from stilus.nodes.boolean import Boolean, false
from stilus.nodes.call import Call
from stilus.nodes.color import RGBA
from stilus.nodes.expression import Expression
from stilus.nodes.function import Function
from stilus.nodes.group import Group
from stilus.nodes.ident import Ident
from stilus.nodes.import_node import Import
from stilus.nodes.literal import Literal
from stilus.nodes.node import CoercionError
from stilus.nodes.null import null
from stilus.nodes.object_node import ObjectNode
from stilus.nodes.return_node import ReturnNode
from stilus.nodes.root import Root
from stilus.nodes.string import String
from stilus.nodes.unit import Unit
from stilus.parser import Parser
from stilus.stack.frame import Frame
from stilus.stack.scope import Scope
from stilus.stack.stack import Stack
from stilus.units import units
from stilus.utils import unwrap
from stilus.visitor.visitor import Visitor

log = logging.getLogger(__name__)


class Evaluator(Visitor):
    def __init__(self, root, parser, options: dict):
        super().__init__(root)
        self.bifs = bifs
        self.parser = parser
        self.options = options
        self.functions = options.get("functions", {})
        # making sure that newly defined functions are builtins
        # for name, function in self.functions.items():
        #     if name not in self.bifs.keys():
        #         log.debug(f'Adding extra built-in function: {name}.')
        #         print(f'Adding extra built-in function: {name}: {function}.')
        #         raw_bifs.append(name)
        #         self.bifs[name] = function
        self.stack = Stack()
        self.imports = options.get("imports", [])
        self.commons = options.get("globals", {})
        self.paths = options.get("paths", [])
        self.prefix = options.get("prefix", "")
        self.filename = options.get("filename", None)
        self.include_css = options.get("include css", False)

        self.resolve_url = False
        # todo: why is this done separately from the bifs?
        if "url" in self.functions:
            url = self.functions["url"]
            if url.__name__ == "resolver" and hasattr(url, "options"):
                log.debug(f"Using this resolver: {url}.")
                self.resolve_url = True

        filename = Path(options.get("filename", "."))
        self.paths.append(str(filename.parent))

        self.common = Frame(root)
        self.stack.append(self.common)

        self.warnings = options.get("warn", False)
        self.calling = []  # todo: remove, use stack
        self.import_stack = []
        self.require_history = {}
        self.result = 0
        self.current_scope = None
        self.ignore_colors = None
        self.property = None

    def vendors(self):
        return [node.string for node in self.lookup("vendors").nodes]

    def import_file(self, node: Import, file, literal, lineno=1, column=1):
        log.debug(f"importing {file}; {self.import_stack}")

        # handling 'require'
        if node.once:
            if self.require_history.get(file, False):
                return null
            self.require_history[file] = True

            if literal and not self.include_css:
                return node

        # avoid overflows from reimporting the same file
        if file in self.import_stack:
            raise ImportError("import loop has been found")

        with open(file, "r") as f:
            source = f.read()

        # shortcut for empty files
        if not source.strip():
            return null

        # expose imports
        node.path = file
        node.dirname = Path(file).parent
        # store modified time
        node.mtime = os.stat(file).st_mtime
        self.paths.append(str(node.dirname))

        if "_imports" in self.options:
            self.options["_imports"].append(node.clone())

        # parse the file
        self.import_stack.append(file)
        filename_node.current_filename = file

        if literal:
            re.sub("\n\n?", "\n", source)
            literal = Literal(
                source, lineno=self.parser.lineno, column=self.parser.column
            )
            literal.lineno = 1
            literal.column = 1
            if not self.resolve_url:
                return literal

        # create block
        block = Block(None, None, lineno=lineno, column=column)
        block.filename = file

        # parse
        merged = {}
        merged.update(self.options)
        merged.update({"root": block})
        parser = Parser(source, merged)

        try:
            block = parser.parse()
        except Exception:
            line = parser.lexer.lineno
            column = parser.lexer.column
            if literal and self.include_css and self.resolve_url:
                self.warn(
                    f"ParseError: {file}:{line}:{column}. "
                    f"This file is included as-is"
                )
                return literal
            else:
                raise ParseError(
                    "Issue when parsing an imported file",
                    filename=file,
                    lineno=line,
                    column=column,
                    input=source,
                )

        # evaluate imported 'root'
        block = block.clone(self.get_current_block())
        block.parent = self.get_current_block()
        block.scope = False
        ret = self.visit(block)
        self.import_stack.pop()

        if not self.resolve_url:  # or not self.resolve_url no_check:
            self.paths.pop()

        return ret

    def visit(self, node):
        try:
            return super().visit(node)
        except StilusError as se:
            if not se.filename:
                raise se
            try:
                with open(str(se.filename)) as f:
                    input = f.read()
            except (AttributeError, FileNotFoundError):
                pass
            raise StilusError(
                self.stack,
                filename=node.filename,
                lineno=node.lineno,
                column=node.column,
                input=input,
            )

    def setup(self):
        self.populate_global_scope()
        for file in reversed(self.imports):
            expr = Expression()
            expr.append(
                String(
                    file, lineno=self.parser.lineno, column=self.parser.column
                )
            )
            self.root.nodes.insert(
                0,
                Import(
                    expr, lineno=self.parser.lineno, column=self.parser.column
                ),
            )

    def populate_global_scope(self):
        """
        Populate the global scope with:
            - css colors
            - user-defined globals
        :return:
        """
        self.common._scope = Scope()

        # colors
        for color, value in colors.items():
            rgba = RGBA(value[0], value[1], value[2], value[3])
            ident = Ident(color, rgba)
            rgba.name = color
            self.common.scope().add(ident)

        # todo: should this be here?
        # self.common.scope().add(Ident('embedurl',
        #                               Function('embedurl', url.fn,
        #                                        lineno=self.parser.lineno,
        #                                        column=self.parser.column)))

        # user defined globals
        commons = self.commons
        for common, val in commons.items():
            if val.name:
                self.common.scope().add(Ident(common, val))

    def evaluate(self):
        self.setup()
        return self.visit(self.root)

    def visit_group(self, group: Group):
        new_nodes = []
        for selector in group.nodes:
            selector.value = self.interpolate(selector)
            new_nodes.append(selector)
        group.nodes = new_nodes
        group.block = self.visit(group.block)
        return group

    def visit_returnnode(self, ret):
        ret.expression = self.visit(ret.expression)
        raise ret

    def visit_media(self, media):
        media.block = self.visit(media.block)
        media.value = self.visit(media.value)
        return media

    def visit_querylist(self, queries):
        for node in queries.nodes:
            self.visit(node)

        if len(queries.nodes) == 1:
            query = queries.nodes[0]
            val = self.lookup(query.type)
            if val:
                if hasattr(val, "first") and hasattr(val.first(), "string"):
                    val = val.first().string
                else:
                    return queries
                parser = Parser(val, self.options)
                queries = self.visit(parser.queries())

        return queries

    def visit_query(self, node):
        node.predicate = self.visit(node.predicate)
        node.type = self.visit(node.type)
        for n in node.nodes:
            self.visit(n)
        return node

    def visit_feature(self, node):
        node.name = self.interpolate(node)
        if node.expr:
            self.result += 1
            node.expr = self.visit(node.expr)
            self.result -= 1
        return node

    def visit_objectnode(self, obj: ObjectNode):
        for key in obj.values.keys():
            obj.values[key] = self.visit(obj.values[key])
        return obj

    def visit_member(self, node):
        left = node.left
        right = node.right
        obj = self.visit(left).first()

        if "objectnode" != obj.node_name:
            raise ParseError(f"{left} has no property .{right}")

        if node.value:
            self.result += 1
            obj.set(right.name, self.visit(node.value))
            self.result -= 1

        return obj.get(right.name)

    def visit_keyframes(self, keyframes):
        if keyframes.fabricated:
            return keyframes
        keyframes.value = self.interpolate(keyframes).strip()
        val = self.lookup(keyframes.value)
        if val:
            if type(val.first()) == Function:
                keyframes.value = val.first().function_name
            else:
                keyframes.value = val.first().string
        keyframes.block = self.visit(keyframes.block)

        if "official" != keyframes.prefix:
            return keyframes

        for prefix in self.vendors():
            # IE never had prefixes for keyframes
            if "ms" == prefix:
                continue
            node = keyframes.clone()
            node.value = keyframes.value
            node.prefix = prefix
            node.block = keyframes.block
            node.fabricated = True
            self.get_current_block().append(node)

        return null

    def visit_function(self, fn, args=None, content=None):
        # check local
        local = self.stack.current_frame().scope().lookup(fn.function_name)
        if local:
            self.warn(
                f'local {local.node_name} "fn.function_name" '
                f"previously defined in this scope"
            )

        # user-defined
        user = self.functions.get(fn.function_name, None)
        if user:
            self.warn(
                f'user-defined function "{fn.function_name}" '
                f"is already defined"
            )

        if fn.function_name in self.bifs:
            self.warn(
                f'built-in function "{fn.function_name}" '
                f"is already defined"
            )

        return fn

    def visit_each(self, each):
        self.result += 1
        expr = utils.unwrap(self.visit(each.expr))
        length = len(expr.nodes)
        val = Ident(each.value)
        key = Ident("__index__")
        if each.key:
            key = Ident(each.key)
        scope = self.get_current_scope()
        block = self.get_current_block()
        vals = []
        self.result -= 1

        each.block.scope = False

        def visit_body(key, value):
            scope.add(value)
            scope.add(key)
            body = self.visit(each.block.clone())
            vals.extend(body.nodes)

        # for prop in obj
        if length == 1 and "objectnode" == expr.nodes[0].node_name:
            obj = expr.nodes[0]
            for prop in obj.values:
                val.value = String(
                    prop, lineno=self.parser.lineno, column=self.parser.column
                )
                key.value = obj.get(prop)  # checkme: works?
                visit_body(key, val)
        else:
            for i, n in enumerate(expr.nodes):
                val.value = n
                key.value = Unit(i)
                visit_body(key, val)

        self.mixin(vals, block)

        if vals and len(vals) > 0:
            return vals[len(vals) - 1]
        else:
            return null

    def visit_call(self, call):
        fn = self.lookup(call.function_name)
        log.debug(f"Visiting {fn}; {call}...")
        if hasattr(fn, "param") and fn.param:
            log.debug(fn.params)

        # url()
        self.ignore_colors = "url" == call.function_name

        # variable function
        if fn and "expression" == fn.node_name:
            fn = fn.nodes[0]

        # not a function? try user-defined or built-ins
        if fn and "function" != fn.node_name:
            fn = self.lookup_function(call.function_name)

        # undefined function? render literal css
        if fn is None or fn.node_name != "function":
            ret = None
            if "calc" == self.unvendorize(call.function_name):
                if call.args.nodes:
                    if call.args.nodes[0]:
                        ret = Literal(
                            call.function_name + str(call.args.nodes[0]),
                            lineno=self.parser.lineno,
                            column=self.parser.column,
                        )
            else:
                ret = self.literal_call(call)
            self.ignore_colors = False
            return ret

        self.calling.append(call.function_name)

        # massive stack
        if len(self.calling) > 200:
            raise ParseError("Maximum stylus call stack size exceeded")

        # first node in expression
        if fn and "expression" == fn.function_name:
            fn = fn.first()

        # evaluate arguments
        self.result += 1
        args = self.visit(call.args)
        if hasattr(args, "map"):
            for key in args.map:
                args.map[key] = self.visit(args.map[key].clone())
        self.result -= 1

        log.debug(
            f"Going to invoke function: {fn} " f"(builtin: {fn.builtin})."
        )
        if fn.builtin or (
            fn.function_name == "url"
            and hasattr(fn, "params")
            and fn.params.__name__ == "resolver"
        ):
            log.debug(f"{fn} is a built-in method ({fn.params}).")
            ret = self.invoke_builtin(fn.params, args)
        elif "function" == fn.node_name:  # user-defined
            # evaluate mixin block
            log.debug(f"{fn} is a regular method.")
            if call.block:
                log.debug(f"{fn} is a mixin.")
                call.block = self.visit(call.block)
            ret = self.invoke_function(fn, args, call.block)

        self.calling.pop()
        self.ignore_colors = False

        return ret

    def visit_ident(self, ident):
        if ident.property:
            # property lookup
            prop = self.lookup_property(ident.name)
            if prop:
                return self.visit(prop.expr.clone())
            return null
        elif ident.value.node_name == "null":
            # lookup
            val = self.lookup(ident.name)
            # object or block mixin
            if val is not None and ident.mixin:
                self.mixin_node(val)
            if val is not None:
                return self.visit(val)
            else:
                return ident
            # return self.visit(val) if val is not None else ident
        else:
            # assign
            self.result += 1
            ident.value = self.visit(ident.value)
            self.result -= 1
            self.get_current_scope().add(ident)
            return ident.value

    def visit_binop(self, binop):
        # special case 'is defined' pseudo binop
        if "is defined" == binop.op:
            return self.is_defined(binop.left)

        self.result += 1
        # visit operands
        op = binop.op
        left = self.visit(binop.left)
        if "||" == op or "&&" == op:
            right = binop.right
        else:
            right = self.visit(binop.right)

        # hack (sic): ternary
        if binop.value:
            value = self.visit(binop.value)
        else:
            value = null

        self.result -= 1

        # operate
        try:
            return self.visit(left.operate(op, right, value))
        except CoercionError as e:
            # disregard coercion issues in equality
            # checks, and simply return false
            if op == "==":
                return Boolean(False)
            elif op == "!=":
                return Boolean(True)
            raise e

    def visit_unaryop(self, unary):
        op = unary.op
        node = self.visit(unary.expr)

        if "!" != op:
            node = node.first().clone()
            utils.assert_type(node, "unit")

        if op == "-":
            node.value = -node.value
        elif op == "+":
            node.value = +node.value
        elif op == "~":
            if isinstance(node.value, float):
                node.value = ~int(node.value)
            else:
                node.value = ~node.value
        elif op == "!":
            return node.to_boolean().negate()

        return node

    def visit_ternary(self, ternary):
        ok = self.visit(ternary.cond).to_boolean()
        if ok.is_true():
            return self.visit(ternary.true_expr)
        else:
            return self.visit(ternary.false_expr)

    def visit_expression(self, expr):
        expr.nodes = [self.visit(node) for node in expr.nodes]

        # support (n * 5)px etc
        if self.castable(expr):
            expr = self.cast(expr)

        return expr

    def visit_arguments(self, args):
        return self.visit_expression(args)

    def visit_property(self, prop):
        name = self.interpolate(prop)
        fn = self.lookup(name)
        call = fn and "function" == fn.first().node_name
        literal = name in self.calling
        _prop = self.property

        prop_lit = True if hasattr(prop, "literal") and prop.literal else False
        if call and not literal and not prop_lit:
            # function of the same node_name
            clone = prop.expr.clone(None, None)
            args = Arguments.from_expression(utils.unwrap(clone))
            prop.name = name
            self.property = prop
            self.result += 1
            self.property.expr = self.visit(prop.expr)
            self.result -= 1
            ret = self.visit(
                Call(
                    name,
                    args,
                    lineno=self.parser.lineno,
                    column=self.parser.column,
                )
            )
            self.property = _prop
            return ret
        else:
            # regular property
            self.result += 1
            prop.name = name
            prop.literal = True
            self.property = prop
            prop.expr = self.visit(prop.expr)
            self.property = _prop
            self.result -= 1
            return prop

    def visit_root(self, block):
        if block != self.root:
            # normalize cached imports
            return self.visit(block.to_block())

        # for i, node in enumerate(list(block.nodes)):
        i = 0
        while i < len(block.nodes):
            block.index = i
            v = self.visit(block.nodes[i])
            if hasattr(block, "mixin") and block.mixin:
                # log.debug(f'Not adding mixin [{v}]!')
                block.mixin = False
            else:
                block.nodes[i] = v
            i += 1

        return block

    def visit_block(self, block):
        self.stack.append(Frame(block))

        index = 0
        while index < len(block.nodes):
            block.index = index
            try:
                v = self.visit(block.nodes[block.index])
                if block.mixin:
                    log.debug(f"Not adding mixin [{v}]!")
                    block.mixin = False
                else:
                    try:
                        block.nodes[index] = v
                    except TypeError:
                        pass

            except ReturnNode as rn:
                if self.result:
                    self.stack.pop()
                    raise rn
                else:
                    block.nodes[block.index] = rn
                    break

            index = block.index + 1
            block.index += 1

        self.stack.pop()
        return block

    def visit_atblock(self, atblock):
        atblock.block = self.visit(atblock.block)
        return atblock

    def visit_atrule(self, atrule):
        atrule.value = self.interpolate(atrule)
        if atrule.block:
            atrule.block = self.visit(atrule.block)
        return atrule

    def visit_supports(self, node):
        condition = node.condition
        self.result += 1
        node.condition = self.visit(condition)
        self.result -= 1
        value = condition.first()
        if len(condition.nodes) == 1 and value.node_name == "string":
            node.condition = value.string
        node.block = self.visit(node.block)
        return node

    def visit_if(self, node):
        block = self.get_current_block()
        negate = node.negate
        self.result += 1
        ok = self.visit(node.cond).first().to_boolean()
        self.result -= 1

        node.block.scope = (
            hasattr(node.block, "has_media") and node.block.has_media()
        )

        # evaluate body
        ret = None
        if negate:
            # unless
            if ok.is_false():
                ret = self.visit(node.block)
        else:
            # if
            if ok.is_true():
                ret = self.visit(node.block)
            # else
            elif node.elses:
                for b in node.elses:
                    # else if
                    if hasattr(b, "cond") and b.cond:
                        b.block.scope = b.block.has_media()
                        self.result += 1
                        cond = self.visit(b.cond).first().to_boolean()
                        self.result -= 1
                        if cond.is_true():
                            ret = self.visit(b.block)
                            break
                    # else:
                    else:
                        b.scope = b.has_media()
                        ret = self.visit(b)

        # mixin conditional statements within
        # a selector group or at-rule
        # fixme: make this pythonistic!
        if (
            ret
            and node.postfix is None
            and hasattr(block, "node")
            and block.node
            and block.node.node_name
            in ["group", "atrule", "media", "supports", "keyframes"]
        ):
            self.mixin(ret.nodes, block)
            return null

        if ret:
            return ret
        return null

    def visit_extend(self, extend, id=None):
        block = self.get_current_block()
        if block.node.node_name != "group":
            block = self.closest_group()
        for selector in extend.selectors:
            c = selector.clone()
            # todo: this is really bad; refactor this
            some_object = {
                "selector": self.interpolate(c).strip(),
                "optional": selector.optional,
                "lineno": c.lineno,
                "column": c.column,
            }
            block.node.extends.append(some_object)
        return null

    def invoke(self, body, stack=None, filename=None):
        if filename:
            self.paths.append(str(Path(filename).parent))

        if self.result:
            ret = self.eval(body.nodes)
            if stack:
                self.stack.pop()
        else:
            body = self.visit(body)
            if stack:
                self.stack.pop()
            self.mixin(body.nodes, self.get_current_block())
            ret = null

        if filename:
            self.paths.pop()

        return ret

    def mixin(self, nodes, block):
        if len(nodes) == 0:
            return None
        head = block.nodes[: block.index]
        tail = block.nodes[block.index + 1 :]
        self._mixin(nodes, head, block)
        block.index = 0
        block.mixin = True
        head.extend(tail)
        block.nodes = head

    # todo: rewrite this; this is not Python >:-(
    def _mixin(self, items, dest, block):
        for item in items:
            checked = False
            media_passed = False
            if item.node_name == "returnnode":
                return
            elif item.node_name == "block":
                checked = True
                self._mixin(item.nodes, dest, block)
            elif item.node_name == "media":
                # fix link to the parent block
                parent_node = item.block.parent.node
                if parent_node and parent_node.node_name != "call":
                    item.block.parent = block
                media_passed = True
            if media_passed or item.node_name == "property":
                value = None
                if hasattr(item, "expr"):
                    value = item.expr
                # prevent `block` mixin recursion
                if (
                    hasattr(item, "literal")
                    and item.literal
                    and hasattr(value, "first")
                    and value.first().node_name == "block"
                ):
                    value = unwrap(value)
                    value.nodes[0] = Literal(
                        "block",
                        lineno=self.parser.lineno,
                        column=self.parser.column,
                    )
            if not checked:
                dest.append(item)

    def mixin_node(self, node):
        node = self.visit(node.first())
        if node.node_name == "objectnode":
            self.mixin_object(node)
            return null
        elif node.node_name in ["block", "atblock"]:
            self.mixin(node.nodes, self.get_current_block())
            return null

    def mixin_object(self, object: ObjectNode):
        s = f"$block {object.to_block()}"
        p = Parser(s, utils.merge({"root": Root()}, self.options))
        try:
            block = p.parse()
        except StilusError as e:
            e.filename = self.filename
            e.lineno = p.lexer.lineno
            e.column = p.lexer.column
            e.input = s
            raise e

        block.parent = self.root
        block.scope = False

        ret = self.visit(block)
        values = ret.first().nodes
        for value in values:
            if value.block:
                self.mixin(value.block.nodes, self.get_current_block())
                break

    def eval(self, vals=None):
        if vals is None:
            return null

        def update(node):
            do_visit = True
            skip_next = False
            if node.node_name == "if":
                if node.block.node_name != "block":
                    node = self.visit(node)
                    do_visit = False
                    skip_next = True
            if not skip_next and node.node_name in ["if", "each", "block"]:
                node = self.visit(node)
                if hasattr(node, "nodes") and node.nodes:
                    node = self.eval(node.nodes)
                do_visit = False
            if do_visit:
                node = self.visit(node)
            return node

        try:
            nodes = [update(node) for node in vals]
        except ReturnNode as rn:
            return rn.expression

        return nodes[-1] if nodes else null

    # todo: fix this; this is really bad
    def invoke_builtin(self, fn, args):
        """
        :param fn:
        :param args:
        :return:
        """
        # map arguments to first node
        # providing a nicer js api for
        # built-in functions. Functions may specify that
        # they wish to accept full expressions
        # via raw_bifs
        if hasattr(fn, "__name__") and fn.__name__ in raw_bifs:
            ret = args.nodes
        else:
            ret = []
            # fit in the args to the functions
            sig = inspect.signature(fn)
            i = 0
            # remove the last parameter (the evaluator) first
            keys = [key for key in sig.parameters.keys()][:-1]
            for key in keys:
                param = sig.parameters.get(key)
                # handle the *args parameters
                if param.name == "args":
                    while i < len(args.nodes):
                        ret.append(utils.unwrap(args.nodes[i].first()))
                        i += 1
                # regular parameters
                elif param.name in args.map.keys():
                    # in the map
                    ret.append(args.map[param.name].first())
                else:
                    # then in the nodes
                    all_arguments = []
                    for nodes in args.nodes:
                        all_arguments.extend(unwrap(nodes))
                    if i < len(unwrap(all_arguments)):
                        ret.append(unwrap(all_arguments)[i].first())
                    i += 1
                    # else: assume remaining parameters are not required

        # invoke builtin function
        body = utils.coerce(
            fn(*ret, evaluator=self),
            False,
            lineno=self.parser.lineno,
            column=self.parser.column,
        )

        # Always wrapping allows js functions
        # to return several values with a single
        # Expression node
        expr = Expression()
        expr.append(body)
        body = expr

        return self.invoke(body)

    def visit_import(self, imported):
        literal = False
        self.result += 1

        path = self.visit(imported.path).first()

        node_name = "require" if imported.once else "import"

        self.result -= 1

        log.debug(f"import {path}")

        # url() passed
        if hasattr(path, "function_name") and path.function_name == "url":
            if hasattr(imported, "once") and imported.once:
                raise StilusError("You cannot @require a url")
            return imported

        # ensure string
        if not hasattr(path, "string"):
            raise StilusError(f"@{node_name} string expected")

        name = path = path.string

        # absolute URL or hash
        m = re.match(
            r"(?:url\s*\(\s*)?[\'\"]?(?:#|(?:https?:)?\/\/)", path, re.I
        )
        if m:
            if imported.once:
                raise StilusError("You cannot @require a url")
            return imported

        # literal
        if path.endswith(".css") or '.css"' in path:
            literal = True
            if not imported.once and not self.include_css:
                return imported

        # support optional .styl
        if not literal and not path.endswith(".styl"):
            path += ".styl"

        # lookup
        found = utils.find(path, self.paths, self.filename)
        if not found:
            found = utils.lookup_index(
                name, self.paths, self.filename, parser=self.parser
            )

        # throw if import failed
        if not found:
            raise TypeError(
                f"failed to locate @{node_name} file in {path}"
                f" {self.paths}"
            )

        block = Block(
            None, None, lineno=self.parser.lineno, column=self.parser.column
        )
        for f in found:
            block.append(
                self.import_file(
                    imported,
                    f,
                    literal,
                    lineno=self.parser.lineno,
                    column=self.parser.column,
                )
            )

        return block

    def lookup_property(self, name):
        i = len(self.stack)
        index = self.get_current_block().index
        top = i
        while i > 0:
            i -= 1
            block = self.stack[i].block
            if (
                block
                and hasattr(block, "node")
                and block.node.node_name
                in [
                    "group",
                    "function",
                    "if",
                    "each",
                    "atrule",
                    "media",
                    "atblock",
                    "call",
                ]
            ):
                nodes = block.nodes
                if i + 1 == top:
                    while index > 0:
                        index -= 1
                        if self.property == nodes[index]:
                            continue
                        other = self.interpolate(nodes[index])
                        if name == other:
                            return nodes[index].clone()
                else:
                    length = len(nodes)
                    while length > 0:
                        length -= 1
                        if (
                            "property" != nodes[length].node_name
                            or self.property == nodes[length]
                        ):
                            continue
                        other = self.interpolate(nodes[length])
                        if name == other:
                            return nodes[length].clone()
        return null

    def closest_block(self):
        for stack in reversed(self.stack):
            block = stack.block
            if (
                hasattr(block, "node")
                and block.node
                and block.node.node_name
                in ["group", "keyframes", "atrule", "atblock", "media", "call"]
            ):
                return block
        # todo: create warning when entering here
        return None

    def closest_group(self):
        for s in self.stack:
            b = s.block
            if hasattr(b, "node") and b.node and b.node.node_name == "group":
                return b

    def selector_stack(self):
        stack = []
        for s in self.stack:
            b = s.block
            if hasattr(b, "node") and b.node and b.node.node_name == "group":
                for s in b.node.nodes:
                    if not s.value:
                        s.value = self.interpolate(s)
                stack.append(b.node.nodes)
        return stack

    def property_expression(self, prop, name):
        expr = Expression(lineno=self.parser.lineno, column=self.parser.column)
        value = prop.expr.clone(None)

        # name
        expr.append(
            String(
                prop.name, lineno=self.parser.lineno, column=self.parser.column
            )
        )

        # replace cyclic call with __CALL__
        def replace(node):
            if (
                node.node_name == "call"
                and hasattr(node, "function_name")
                and name == node.function_name
            ):
                return Literal(
                    "__CALL__",
                    lineno=self.parser.lineno,
                    column=self.parser.column,
                )
            if hasattr(node, "nodes") and node.nodes:
                node.nodes = [replace(n) for n in node.nodes]
            return node

        replace(value)
        expr.append(value)
        return expr

    def lookup(self, name):
        """
        Lookup `name`, with support for JavaScript functions, and BIFs.
        :param name:
        :return:
        """
        if self.ignore_colors and name in colors:
            return
        val = self.stack.lookup(name)
        if val is not None:  # fixme: implement __len__()
            return utils.unwrap(val)
        else:
            return self.lookup_function(name)

    def interpolate(self, node):
        is_selector = "selector" == node.node_name

        def to_string(node):
            if node.node_name == "ident":
                return node.name
            elif node.node_name == "function":
                return node.function_name
            elif node.node_name in ["literal", "string"]:
                if (
                    self.prefix
                    and not node.prefixed
                    and not hasattr(node.value, "node_name")
                ):
                    node.value = re.sub(r"\.", f".{self.prefix}", node.value)
                    node.prefixed = True
                return node.value
            elif node.node_name == "unit":
                # interpolation inside keyframes
                if "%" == node.type:
                    return f"{int(node.value)}%"
                else:
                    return f"{int(node.value)}"
            elif node.node_name == "member":
                return to_string(self.visit(node))
            elif node.node_name == "expression":
                # prevent cyclic 'selector()' calls
                if (
                    self.calling
                    and "selector" in self.calling
                    and self._selector
                ):
                    return self._selector
                self.result += 1
                ret = to_string(self.visit(node).first())
                self.result -= 1
                if is_selector:
                    self._selector = ret
                return ret

        if node and hasattr(node, "segments"):
            s = ""
            for segment in node.segments:
                s += str(to_string(segment))
            return s
            # return ''.join(map(to_string, node.segments))
        else:
            return to_string(node)

    def lookup_function(self, name):
        function = self.functions.get(name, self.bifs.get(name, None))
        if function:
            log.debug(f"Lookup function for {name} " f"returned: {function}.")
            return Function(
                name,
                function,
                lineno=self.parser.lineno,
                column=self.parser.column,
            )
        else:
            log.debug(f"No function found for {name}.")
            return None

    def is_defined(self, node):
        if "ident" == node.node_name:
            return Boolean(self.lookup(node.name))
        else:
            raise ParseError(
                f'invalid "is defined" ' f"check on non-variable {node}"
            )

    def cast(self, expr: Expression):
        return Unit(expr.first().value, expr.nodes[1].name)

    def castable(self, expr: Expression):
        return (
            len(expr.nodes) == 2
            and expr.first().node_name == "unit"
            and expr.nodes[1]
            and hasattr(expr.nodes[1], "name")
            and expr.nodes[1].name in units
        )

    def warn(self, message):
        if not self.warnings:
            return
        msg = f"Warning: {message}"
        print(msg)
        log.info(msg)

    def get_current_scope(self):
        return self.stack.current_frame().scope()

    def get_current_frame(self):
        return self.stack.current_frame()

    def get_current_block(self):
        current_frame = self.get_current_frame()
        if current_frame:
            b = current_frame.block
            return b
        return None

    def set_current_block(self, block):
        current_frame = self.get_current_frame()
        if current_frame:
            current_frame.block = block
        else:
            raise TypeError

    def unvendorize(self, prop: str):
        for vendor in self.vendors():
            if vendor != "official":
                vendor = f"-{vendor}-"
            if vendor in prop:
                return prop.replace(vendor, "")
        return prop

    def literal_call(self, call):
        call.args = self.visit(call.args)
        return call

    def invoke_function(self, fn, args, content):
        block = Block(
            fn.block.parent,
            None,
            lineno=self.parser.lineno,
            column=self.parser.column,
        )

        # clone the function body to prevent mutation of subsequent calls
        body = fn.block.clone(block)

        mixin_block = self.stack.current_frame().block

        # new block scope
        self.stack.append(Frame(block))
        scope = self.get_current_scope()

        # normalize arguments
        if args.node_name != "arguments":
            expr = Expression()
            expr.append(args)
            args = Arguments.from_expression(expr)

        # arguments
        scope.add(Ident("arguments", args))

        # mixin scope introspection
        bn = mixin_block.node_name
        if self.result:
            scope.add(Ident("mixin", false))
        else:
            scope.add(
                Ident(
                    "mixin",
                    String(
                        bn,
                        lineno=self.parser.lineno,
                        column=self.parser.column,
                    ),
                )
            )

        # current property
        if self.property:
            prop = self.property_expression(self.property, fn.function_name)
            scope.add(
                Ident(
                    "current-property",
                    prop,
                    lineno=self.parser.lineno,
                    column=self.parser.column,
                )
            )
        else:
            scope.add(
                Ident(
                    "current-property",
                    null,
                    lineno=self.parser.lineno,
                    column=self.parser.column,
                )
            )

        # current call stack
        expr = Expression(lineno=self.parser.lineno, column=self.parser.column)
        for call in reversed(self.calling[:-1]):
            expr.append(
                Literal(
                    call, lineno=self.parser.lineno, column=self.parser.column
                )
            )
        scope.add(
            Ident(
                "called-from",
                expr,
                lineno=self.parser.lineno,
                column=self.parser.column,
            )
        )

        # inject arguments as locals
        # todo: rewrite this!
        i = 0
        for index, node in enumerate(fn.params.nodes):
            # rest param support
            if node.rest:
                node.value = Expression(
                    lineno=self.parser.lineno, column=self.parser.column
                )
                for n in args.nodes[i:]:
                    node.value.append(n)
                node.value.preserve = True
                node.value.is_list = args.is_list
            else:
                # argument default support

                # in dict?
                arg = args.map.get(node.name)
                # next node?
                if not arg:
                    if hasattr(args, "nodes") and i < len(args.nodes):
                        arg = args.nodes[i]
                    i += 1

                node = node.clone()
                if arg is not None:
                    if hasattr(arg, "is_empty") and arg.is_empty():
                        args.nodes[i - 1] = self.visit(node)
                    else:
                        node.value = arg
                else:
                    args.append(node.value)

                # required argument not satisfied
                if not node.value:
                    raise TypeError(f'argument "{node}" required for {fn}')

            scope.add(node)

        # mixin block
        if content:
            scope.add(Ident("block", content, True))

        # invoke
        return self.invoke(body, True, fn.filename)



from stilus.nodes.color import RGBA
from stilus.nodes.unit import Unit
from stilus.utils import assert_color, assert_type


def best(top, bottom):
    divisor = 255 if 0 < top - bottom else 0
    return (top - bottom) / (divisor - bottom)


def best_alpha(top, bottom):
    values = []
    values.append(best(top.r, bottom.r))
    values.append(best(top.g, bottom.g))
    values.append(best(top.b, bottom.b))
    return max(values)


def transparentify(top, bottom=None, alpha=None, evaluator=None):
    if isinstance(bottom, Unit):
        # dirty hack
        alpha = bottom
        bottom = None
    assert_color(top)
    top = top.rgba()
    if not bottom:
        bottom = RGBA(255, 255, 255, 1)
    if not alpha and bottom and hasattr(bottom, "rgba") and not bottom.rgba:
        alpha = bottom
        bottom = RGBA(255, 255, 255, 1)
    assert_color(bottom)
    bottom = bottom.rgba()

    ba = best_alpha(top, bottom)

    if alpha:
        assert_type(alpha, "unit", "aplha")
        if alpha.type == "%":
            ba = alpha.value / 100
        elif not alpha.type:
            ba = alpha = alpha.value

    ba = max(min(ba, 1), 0)

    if ba == 0:
        return RGBA(bottom.r, bottom.g, bottom.b, round(ba * 100) / 100)
    else:
        return RGBA(
            bottom.r + (top.r - bottom.r) / ba,
            bottom.g + (top.g - bottom.g) / ba,
            bottom.b + (top.b - bottom.b) / ba,
            round(ba * 100) / 100,
        )



from stilus import utils
from stilus.nodes.ident import Ident
from stilus.nodes.string import String


def slice(values, start, end=None, evaluator=None):
    start = int(start.nodes[0].value)
    if end and len(end.nodes) > 0:
        end = int(end.nodes[0].value)

    values = utils.unwrap(values).nodes

    if len(values) > 1:
        if end:
            return utils.coerce(values[start:end], raw=True)
        else:
            return utils.coerce(values[start:], raw=True)

    if end:
        result = values[0].string[start:end]
    else:
        result = values[0].string[start:]

    if isinstance(values[0], Ident):
        return Ident(result)
    else:
        return String(result)



import json

from .atrule import Atrule


class Keyframes(Atrule):
    def __init__(self, segments, prefix="official", lineno=1, column=1):
        super().__init__("keyframes", lineno=lineno, column=column)
        self.segments = segments
        if prefix:
            self.prefix = prefix
        else:
            self.prefix = "official"
        self.frames = None
        self.fabricated = False
        self.block = None

    def __str__(self):
        return f'@keyframes {"".join(str(self.segments))}'

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return self.segments, self.prefix

    def __eq__(self, other):
        if isinstance(other, Keyframes):
            return self.__key() == other.__key()
        return False

    def clone(self, parent=None, node=None):
        clone = Keyframes([], lineno=self.lineno, column=self.column)
        clone.filename = self.filename
        clone.segments = [segment for segment in self.segments]
        clone.prefix = self.prefix
        clone.block = self.block.clone(parent, clone)
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Keyframes",
                "segments": self.segments,
                "prefix": self.prefix,
                "block": self.block,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



from stilus.utils import assert_string
from stilus.visitor.normalizer import Normalizer


def selector_exists(selector, evaluator=None):
    assert_string(selector, "selector")

    visitor = Normalizer(
        evaluator.root.clone(), evaluator.parser, evaluator.options
    )
    visitor.visit(visitor.root)

    return selector.string in visitor.selector_map



import math as m

from stilus.nodes.string import String
from stilus.nodes.unit import Unit


# todo: convert all javascript math functions to python and vice versa
def math_prop(prop: String, evaluator=None):
    """Get Math ``prop``"""
    if hasattr(m, prop.string):
        return Unit(getattr(m, prop.string))
    elif hasattr(m, str(prop.string.upper())):
        upper = prop.string.upper()
        return Unit(getattr(m, upper))
    elif hasattr(m, str(prop.string.lower())):
        lower = prop.string.lower()
        return Unit(getattr(m, lower))
    else:
        raise ValueError(f"Oops; could not return math.{prop}.")



from stilus.nodes.expression import Expression
from stilus.nodes.ident import Ident
from stilus.nodes.string import String
from stilus.utils import assert_string


def split(delim, value, evaluator=None):
    assert_string(delim, "delimiter")
    assert_string(value, "val")
    words = value.string.split(delim.string)
    expr = Expression()
    for word in words:
        if isinstance(value, Ident):
            addition = Ident(word)
        else:
            addition = String(word)
        expr.append(Ident(addition))
    return expr



from stilus.functions.blend import blend
from stilus.functions.luminosity import luminosity
from stilus.nodes.color import RGBA, Color
from stilus.nodes.literal import Literal
from stilus.nodes.null import Null
from stilus.nodes.object_node import ObjectNode
from stilus.nodes.unit import Unit
from stilus.utils import assert_color, stilus_round


def contrast(top: Color = None, bottom: Color = None, evaluator=None):
    if not isinstance(top, Color) and not isinstance(bottom, Color):
        c = "" if isinstance(top, (Null, type(None))) else f"{top}"
        return Literal(f"contrast({c})")
    result = ObjectNode()
    if not bottom:
        bottom = RGBA(255, 255, 255, 1)
    assert_color(bottom)
    bottom = bottom.rgba()

    def contrast_function(top, bottom):
        if 1 > top.a:
            top = blend(top, bottom)
        l1 = luminosity(bottom).value + 0.05
        l2 = luminosity(top).value + 0.05
        ratio = l1 / l2

        if l2 > l1:
            ratio = 1 / ratio

        return round(ratio * 10) / 10

    if 1 <= bottom.a:
        result_ratio = Unit(contrast_function(top, bottom))
        result.set("ratio", result_ratio)
        result.set("error", Unit(0))
        result.set("min", result_ratio)
        result.set("max", result_ratio)
    else:
        on_black = contrast_function(top, blend(bottom, RGBA(0, 0, 0, 1)))
        on_white = contrast_function(
            top, blend(bottom, RGBA(255, 255, 255, 1))
        )
        the_max = max(on_black, on_white)

        def process_channel(top_channel, bottm_channel):
            return min(
                max(
                    0,
                    (top_channel - bottm_channel * bottom.a) / (1 - bottom.a),
                ),
                255,
            )

        closest = RGBA(
            process_channel(top.r, bottom.r),
            process_channel(top.g, bottom.g),
            process_channel(top.b, bottom.b),
            1,
        )

        the_min = contrast_function(top, blend(bottom, closest))

        result.set("ratio", Unit(stilus_round((the_min + the_max) * 50) / 100))
        result.set("error", Unit(stilus_round((the_max - the_min) * 50) / 100))
        result.set("min", Unit(the_min))
        result.set("max", Unit(the_max))

    return result



from stilus.nodes.expression import Expression
from stilus.nodes.string import String
from stilus.selector_parser import SelectorParser


def selectors(evaluator=None):
    stack = evaluator.selector_stack()
    expr = Expression(is_list=True)
    if stack:
        for i, group in enumerate(stack):
            if len(group) > 1:
                selector_values = []
                for selector in group:
                    nested = SelectorParser(selector.value).parse()["nested"]
                    if nested and i != 0:
                        selector_values.append(f"& {selector.value}")
                    else:
                        selector_values.append(f"{selector.value}")
                expr.append(String(",".join(selector_values)))
            else:
                selector = group[0].value
                nested = SelectorParser(selector).parse()["nested"]
                if nested and i != 0:
                    expr.append(String(f"& {selector}"))
                else:
                    expr.append(String(f"{selector}"))
    else:
        expr.append(String("&"))
    return expr



class ParseError(Exception):
    def __init__(
        self, message, filename=None, lineno=None, column=None, input=None
    ):
        super().__init__(message)
        self.message = message
        self.filename = filename
        self.lineno = lineno
        self.column = column
        self.input = input


class StilusError(Exception):
    def __init__(
        self, message, filename=None, lineno=None, column=None, input=None
    ):
        super().__init__(message)
        self.message = message
        self.filename = filename
        self.lineno = lineno
        self.column = column
        self.input = input
        self.from_stilus = None

    def __str__(self):
        if self.message:
            return f"Could not compile: {self.message}"
        else:
            return f"Could not compile."



import json

from .node import Node


class Atrule(Node):
    def __init__(self, type, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.type = type
        self.block = None
        self.segments = None

    def __str__(self):
        return f"@{self.type}"

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, Atrule):
            return self.__key() == other.__key()
        return False

    def has_output(self):
        return self.block and self._has_output(self.block)

    def _has_output(self, block):
        # only placeholder selectors
        if hasattr(block, "nodes") and all(
            node.node_name == "group" and node.has_only_placeholders()
            for node in block.nodes
        ):
            return False

        # something visible
        return hasattr(block, "nodes") and any(
            node.node_name in ["property", "literal", "import"]
            or self._has_output(node)
            or (hasattr(node, "block") and self._has_output(node.block))
            for node in block.nodes
        )

    def has_only_properties(self):
        if not self.block:
            return False
        for node in self.block.nodes:
            if node.node_name in ["property", "expression", "comment"]:
                continue
            else:
                return False
        return True

    def clone(self, parent=None, node=None):
        clone = Atrule(self.type)
        if self.block:
            clone.block = self.block.clone(parent, clone)
        clone.segments = [node.clone(parent, clone) for node in self.segments]
        clone.lineno = self.lineno
        clone.column = self.column
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Atrule",
                "type": self.type,
                "segments": self.segments,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



import json

from .node import Node


class Ident(Node):
    def __init__(self, name, value=None, mixin=False, lineno=1, column=1):
        super().__init__(value, lineno=lineno, column=column)
        self.name = name
        self.string = name
        # todo: clean this up
        from stilus.nodes.null import null

        self.value = null if value is None else value
        self.mixin = mixin
        self.property = None
        self.rest = False

    def __str__(self):
        return f"{self.string}"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return (
            self.name,
            self.value,
            self.mixin,
            self.property,
            self.lineno,
            self.column,
        )

    def __eq__(self, other):
        if isinstance(other, Ident):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.name)

    def hash(self):
        return self.name

    def clone(self, parent=None, node=None):
        clone = Ident(self.name)
        clone.value = self.value.clone(parent, clone)
        clone.mixin = self.mixin
        clone.lineno = self.lineno
        clone.column = self.column
        clone.filename = self.filename
        clone.property = self.property
        clone.rest = self.rest
        return clone

    def is_empty(self):
        return self.value is None

    def to_json(self):
        return json.dumps(
            {
                "__type": "Ident",
                "node_name": self.node_name,
                "val": self.value,
                "mixin": self.mixin,
                "property": self.property,
                "rest": self.rest,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )

    def coerce(self, other):
        if other.node_name in ["ident", "string", "literal"]:
            return Ident(other.string)
        elif other.node_name == "unit":
            return Ident(other.__str__())
        else:
            return super().coerce(other)

    def operate(self, op, right, value=None):
        value = right.first()
        if op == "-":
            if value.node_name == "unit":
                from stilus.nodes.expression import Expression

                expression = Expression()
                value = value.clone()
                value.value = -value.value
                expression.append(self)
                expression.append(value)
                return expression
        elif op == "+":
            return Ident(self.string + self.coerce(value).string)
        return super().operate(op, right)



import json

from .node import Node


class Each(Node):
    def __init__(self, value, key, expr=None, block=None, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.value = value
        self.key = key
        self.expr = expr
        self.block = block

    def __str__(self):
        return f"{self.value}: {self.key} [{self.expr}] {self.block}"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.node_name, self.value, self.key, self.expr, self.block

    def __eq__(self, other):
        if isinstance(other, Each):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.__key())

    def clone(self, parent=None, node=None):
        clone = Each(self.value, self.key)
        clone.expr = self.expr.clone(parent, clone)
        clone.block = self.block.clone(parent, clone)
        clone.lineno = self.lineno
        clone.column = self.column
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Each",
                "val": self.value,
                "key": self.key,
                "expr": self.expr,
                "block": self.block,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



import json

from .expression import Expression


class Arguments(Expression):
    def __init__(self, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.map = {}
        self.is_list = False

    def __str__(self):
        string_nodes = [str(node) for node in self.nodes]
        if self.is_list:
            return ", ".join(string_nodes)
        else:
            return " ".join(string_nodes)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if isinstance(other, Arguments):
            return self.map == other.map
        return False

    def __hash__(self):
        return hash(self.map)

    @staticmethod
    def from_expression(expression: Expression):
        arguments = Arguments(
            lineno=expression.lineno, column=expression.column
        )
        arguments.is_list = expression.is_list
        for node in expression.nodes:
            arguments.append(node)
        return arguments

    def clone(self, parent=None, node=None):
        clone = Arguments.from_expression(Expression.clone(self, parent))
        clone.is_list = self.is_list
        clone.map = {
            key: value.clone(parent, clone)
            for (key, value) in self.map.items()
        }
        clone.is_list = self.is_list
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Arguments",
                "value": self.value,
                "map": self.map,
                "isList": self.is_list,
                "preserve": self.preserve,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
                "nodes": self.nodes,
            }
        )



import json

from .node import Node


class Ternary(Node):
    def __init__(self, cond, true_expr, false_expr, lineno=1, column=1):
        super().__init__(lineno=lineno, column=1)
        self.cond = cond
        self.true_expr = true_expr
        self.false_expr = false_expr

    def __str__(self):
        return f"({self.cond},{self.true_expr},{self.false_expr})"

    def __repr__(self):
        return [n.__str__() for n in self.nodes]

    def __key(self):
        return self.cond, self.true_expr, self.false_expr

    def __eq__(self, other):
        if isinstance(other, Ternary):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.__key())

    def clone(self, parent=None, node=None):
        clone = Ternary(
            None, None, None, lineno=self.lineno, column=self.column
        )
        clone.cond = self.cond.clone(parent, clone)
        clone.true_expr = self.true_expr.clone(parent, clone)
        clone.false_expr = self.false_expr.clone(parent, clone)
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "UnaryOp",
                "op": self.op,
                "expr": self.expr,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



from pathlib import Path

from stilus.utils import assert_string


def basename(p, ext=None, evaluator=None):
    assert_string(p, "path")
    result = Path(p).name
    if ext and result.endswith(ext.name):
        result[: len(ext.name)]
    return result



from stilus.functions import (
    addProperty,
    adjust,
    alpha,
    baseConvert,
    basename,
    blend,
    blue,
    clone,
    contrast,
    convert,
    currentMedia,
    define,
    dirname,
    error,
    extname,
    green,
    hsl,
    hsla,
    hue,
    imageSize,
    json,
    length,
    lightness,
    listSeparator,
    lookup,
    luminosity,
    match,
    math,
    math_prop,
    merge,
    operate,
    oppositePosition,
    p,
    pathjoin,
    pop,
    prefixClasses,
    push,
    range,
    red,
    remove,
    replace,
    rgb,
    rgba,
    s,
    saturation,
    selector,
    selectorExists,
    selectors,
    shift,
    stilus_slice,
    stilus_split,
    substr,
    tan,
    trace,
    transparentify,
    type_function,
    unit,
    unquote,
    unshift,
    url,
    warn,
)

bifs = {
    "-math-prop": math_prop.math_prop,
    "add-property-function": addProperty.add_property,
    "adjust": adjust.adjust,
    "alpha": alpha.alpha,
    "append": push.push,
    "base-convert": baseConvert.base_convert,
    "basename": basename.basename,
    "blend": blend.blend,
    "blue": blue.blue,
    "clone": clone.clone,
    "convert": convert.convert,
    "contrast": contrast.contrast,
    "current-media": currentMedia.current_media,
    "define": define.define,
    "dirname": dirname.dirname,
    "embedurl": url.url,
    "error": error.error,
    "extend": merge.merge,
    "extname": extname.extname,
    "green": green.green,
    "hsl": hsl.hsl,
    "hsla": hsla.hsla,
    "hue": hue.hue,
    "image-size": imageSize.image_size,
    "json": json.json_function,
    "length": length.length,
    "lightness": lightness.lightness,
    "list-separator": listSeparator.list_separator,
    "luminosity": luminosity.luminosity,
    "lookup": lookup.lookup,
    "match": match.match,
    "math": math.math,
    "merge": merge.merge,
    "operate": operate.operate,
    "opposite-position": oppositePosition.opposite_position,
    "p": p.p,
    "pathjoin": pathjoin.pathjoin,
    "pop": pop.pop,
    "-prefix-classes": prefixClasses.prefix_classes,
    "push": push.push,
    "prepend": unshift.unshift,
    "range": range.range_function,
    "red": red.red,
    "remove": remove.remove,
    "replace": replace.replace,
    "rgb": rgb.rgb,
    "rgba": rgba.rgba,
    "s": s.s,
    "saturation": saturation.saturation,
    "selector": selector.selector,
    "selector-exists": selectorExists.selector_exists,
    "selectors": selectors.selectors,
    "shift": shift.shift,
    "slice": stilus_slice.slice,
    "split": stilus_split.split,
    "substr": substr.substr,
    "tan": tan.tan,
    "trace": trace.trace,
    "transparentify": transparentify.transparentify,
    "type": type_function.type_function,
    "typeof": type_function.type_function,
    "unit": unit.unit,
    "unquote": unquote.unquote,
    "unshift": unshift.unshift,
    "warn": warn.warn,
}

# todo: automate the creation of these (based on the function)
raw_bifs = [
    "add_property",
    "append",
    "base_convert",
    "clone",
    "embedurl",
    "extend",
    "length",
    "list_separator",
    "merge",
    "opposite_position",
    "p",
    "pop",
    "push",
    "prepend",
    "s",
    "selector",
    "shift",
    "slice",
    "unshift",
]

# todo: handle this better
raw_bifs.extend(["url", "resolver"])



import json

from .node import Node


class Nothing(Node):
    """Empty object; used to end the visiting tree."""

    def __init__(self):
        super().__init__()

    def __str__(self):
        return f"nothing"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.node_name

    def __eq__(self, other):
        if isinstance(other, Nothing):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.__key())

    def to_json(self):
        return json.dumps(
            {
                "__type": "Nothing",
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )
